"""Defines the unit tests for the :mod:`colour.plotting.section` module."""

import unittest
from matplotlib.pyplot import Axes, Figure

from colour.geometry import primitive_cube
from colour.models import RGB_COLOURSPACE_sRGB, RGB_to_XYZ
from colour.plotting import (
    plot_visible_spectrum_section,
    plot_RGB_colourspace_section,
)
from colour.plotting.section import (
    plot_hull_section_colours,
    plot_hull_section_contour,
)
from colour.utilities import is_trimesh_installed

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestPlotHullSectionColours",
    "TestPlotHullSectionContour",
    "TestPlotVisibleSpectrumSection",
    "TestPlotRGBColourspaceSection",
]


class TestPlotHullSectionColours(unittest.TestCase):
    """
    Define :func:`colour.plotting.section.plot_hull_section_colours`
    definition unit tests methods.
    """

    def test_plot_hull_section_colours(self):
        """
        Test :func:`colour.plotting.section.plot_hull_section_colours`
        definition.
        """

        if not is_trimesh_installed:  # pragma: no cover
            return

        import trimesh

        vertices, faces, _outline = primitive_cube(1, 1, 1, 64, 64, 64)
        XYZ_vertices = RGB_to_XYZ(
            vertices["position"] + 0.5,
            RGB_COLOURSPACE_sRGB.whitepoint,
            RGB_COLOURSPACE_sRGB.whitepoint,
            RGB_COLOURSPACE_sRGB.matrix_RGB_to_XYZ,
        )
        hull = trimesh.Trimesh(XYZ_vertices, faces, process=False)

        figure, axes = plot_hull_section_colours(hull)

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        figure, axes = plot_hull_section_colours(hull, axis="+x")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        figure, axes = plot_hull_section_colours(hull, axis="+y")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotHullSectionContour(unittest.TestCase):
    """
    Define :func:`colour.plotting.section.plot_hull_section_contour`
    definition unit tests methods.
    """

    def test_plot_hull_section_contour(self):
        """
        Test :func:`colour.plotting.section.plot_hull_section_contour`
        definition.
        """

        if not is_trimesh_installed:  # pragma: no cover
            return

        import trimesh

        vertices, faces, _outline = primitive_cube(1, 1, 1, 64, 64, 64)
        XYZ_vertices = RGB_to_XYZ(
            vertices["position"] + 0.5,
            RGB_COLOURSPACE_sRGB.whitepoint,
            RGB_COLOURSPACE_sRGB.whitepoint,
            RGB_COLOURSPACE_sRGB.matrix_RGB_to_XYZ,
        )
        hull = trimesh.Trimesh(XYZ_vertices, faces, process=False)

        figure, axes = plot_hull_section_contour(hull)

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotVisibleSpectrumSection(unittest.TestCase):
    """
    Define :func:`colour.plotting.section.plot_visible_spectrum_section`
    definition unit tests methods.
    """

    def test_plot_visible_spectrum_section(self):
        """
        Test :func:`colour.plotting.section.plot_visible_spectrum_section`
        definition.
        """

        if not is_trimesh_installed:  # pragma: no cover
            return

        figure, axes = plot_visible_spectrum_section()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotRGBColourspaceSection(unittest.TestCase):
    """
    Define :func:`colour.plotting.section.plot_RGB_colourspace_section`
    definition unit tests methods.
    """

    def test_plot_RGB_colourspace_section(self):
        """
        Test :func:`colour.plotting.section.plot_RGB_colourspace_section`
        definition.
        """

        if not is_trimesh_installed:  # pragma: no cover
            return

        figure, axes = plot_RGB_colourspace_section("sRGB")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.models.rgb.rgb_colourspace` module."""

import numpy as np
import re
import textwrap
import unittest
from itertools import permutations

from colour.models import (
    RGB_COLOURSPACES,
    RGB_Colourspace,
    XYZ_to_RGB,
    RGB_to_XYZ,
    matrix_RGB_to_RGB,
    RGB_to_RGB,
    chromatically_adapted_primaries,
    normalised_primary_matrix,
    eotf_inverse_sRGB,
    eotf_sRGB,
)
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestRGB_Colourspace",
    "TestXYZ_to_RGB",
    "TestRGB_to_XYZ",
    "TestMatrix_RGB_to_RGB",
    "TestRGB_to_RGB",
]


class TestRGB_Colourspace(unittest.TestCase):
    """
    Define :class:`colour.colour.models.RGB_Colourspace` class unit
    tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        p = np.array([0.73470, 0.26530, 0.00000, 1.00000, 0.00010, -0.07700])
        whitepoint = np.array([0.32168, 0.33767])
        matrix_RGB_to_XYZ = np.identity(3)
        matrix_XYZ_to_RGB = np.identity(3)
        self._colourspace = RGB_Colourspace(
            "RGB Colourspace",
            p,
            whitepoint,
            "ACES",
            matrix_RGB_to_XYZ,
            matrix_XYZ_to_RGB,
            lambda x: x,
            lambda x: x,
        )

    def test_required_attributes(self):
        """Test the presence of required attributes."""

        required_attributes = (
            "name",
            "primaries",
            "whitepoint",
            "whitepoint_name",
            "matrix_RGB_to_XYZ",
            "matrix_XYZ_to_RGB",
            "cctf_encoding",
            "cctf_decoding",
            "use_derived_matrix_RGB_to_XYZ",
            "use_derived_matrix_XYZ_to_RGB",
        )

        for attribute in required_attributes:
            self.assertIn(attribute, dir(RGB_Colourspace))

    def test_required_methods(self):
        """Test the presence of required methods."""

        required_methods = (
            "__init__",
            "__str__",
            "__repr__",
            "use_derived_transformation_matrices",
            "chromatically_adapt",
            "copy",
        )

        for method in required_methods:
            self.assertIn(method, dir(RGB_Colourspace))

    def test__str__(self):
        """
        Test :meth:`colour.models.rgb.rgb_colourspace.RGB_Colourspace.__str__`
        method.
        """

        self.assertEqual(
            re.sub(" at 0x\\w+>", "", str(self._colourspace)),
            textwrap.dedent(
                """
    RGB Colourspace
    ---------------

    Primaries          : [[  7.34700000e-01   2.65300000e-01]
                          [  0.00000000e+00   1.00000000e+00]
                          [  1.00000000e-04  -7.70000000e-02]]
    Whitepoint         : [ 0.32168  0.33767]
    Whitepoint Name    : ACES
    Encoding CCTF      : <function TestRGB_Colourspace.setUp.<locals>.<lambda>
    Decoding CCTF      : <function TestRGB_Colourspace.setUp.<locals>.<lambda>
    NPM                : [[ 1.  0.  0.]
                          [ 0.  1.  0.]
                          [ 0.  0.  1.]]
    NPM -1             : [[ 1.  0.  0.]
                          [ 0.  1.  0.]
                          [ 0.  0.  1.]]
    Derived NPM        : [[  9.52552396e-01   0.00000000e+00   9.36786317e-05]
                          [  3.43966450e-01   7.28166097e-01  -7.21325464e-02]
                          [  0.00000000e+00   0.00000000e+00   1.00882518e+00]]
    Derived NPM -1     : [[  1.04981102e+00   0.00000000e+00  -9.74845406e-05]
                          [ -4.95903023e-01   1.37331305e+00   9.82400361e-02]
                          [  0.00000000e+00   0.00000000e+00   9.91252018e-01]]
    Use Derived NPM    : False
    Use Derived NPM -1 : False"""
            )[1:],
        )

    def test__repr__(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.RGB_Colourspace.\
__repr__` method.
        """

        self.assertEqual(
            re.sub(" at 0x\\w+>", "", repr(self._colourspace)),
            textwrap.dedent(
                """
        RGB_Colourspace(RGB Colourspace,
                        [[  7.34700000e-01,   2.65300000e-01],
                         [  0.00000000e+00,   1.00000000e+00],
                         [  1.00000000e-04,  -7.70000000e-02]],
                        [ 0.32168,  0.33767],
                        ACES,
                        [[ 1.,  0.,  0.],
                         [ 0.,  1.,  0.],
                         [ 0.,  0.,  1.]],
                        [[ 1.,  0.,  0.],
                         [ 0.,  1.,  0.],
                         [ 0.,  0.,  1.]],
                        <function TestRGB_Colourspace.setUp.<locals>.<lambda>,
                        <function TestRGB_Colourspace.setUp.<locals>.<lambda>,
                        False,
                        False)"""
            )[1:],
        )

    def test_use_derived_transformation_matrices(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.RGB_Colourspace.\
use_derived_transformation_matrices` method.
        """

        np.testing.assert_array_equal(
            self._colourspace.matrix_RGB_to_XYZ, np.identity(3)
        )
        np.testing.assert_array_equal(
            self._colourspace.matrix_XYZ_to_RGB, np.identity(3)
        )

        self._colourspace.use_derived_transformation_matrices()

        np.testing.assert_almost_equal(
            self._colourspace.matrix_RGB_to_XYZ,
            np.array(
                [
                    [0.95255240, 0.00000000, 0.00009368],
                    [0.34396645, 0.72816610, -0.07213255],
                    [0.00000000, 0.00000000, 1.00882518],
                ]
            ),
            decimal=7,
        )
        np.testing.assert_almost_equal(
            self._colourspace.matrix_XYZ_to_RGB,
            np.array(
                [
                    [1.04981102, 0.00000000, -0.00009748],
                    [-0.49590302, 1.37331305, 0.09824004],
                    [0.00000000, 0.00000000, 0.99125202],
                ]
            ),
            decimal=7,
        )

        self._colourspace.use_derived_matrix_RGB_to_XYZ = False
        np.testing.assert_array_equal(
            self._colourspace.matrix_RGB_to_XYZ, np.identity(3)
        )
        self._colourspace.use_derived_matrix_XYZ_to_RGB = False
        np.testing.assert_array_equal(
            self._colourspace.matrix_XYZ_to_RGB, np.identity(3)
        )

    def test_chromatically_adapt(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.RGB_Colourspace.\
chromatically_adapt` method.
        """

        whitepoint_t = np.array([0.31270, 0.32900])
        colourspace = self._colourspace.chromatically_adapt(
            whitepoint_t, "D50", "Bradford"
        )

        np.testing.assert_array_almost_equal(
            colourspace.primaries,
            np.array(
                [
                    [0.73485524, 0.26422533],
                    [-0.00617091, 1.01131496],
                    [0.01596756, -0.06423550],
                ]
            ),
            decimal=7,
        )
        np.testing.assert_array_almost_equal(
            colourspace.whitepoint, whitepoint_t, decimal=7
        )

        self.assertEqual(colourspace.whitepoint_name, "D50")

        np.testing.assert_array_almost_equal(
            colourspace.primaries,
            chromatically_adapted_primaries(
                self._colourspace.primaries,
                self._colourspace.whitepoint,
                whitepoint_t,
                "Bradford",
            ),
            decimal=7,
        )

        np.testing.assert_array_almost_equal(
            colourspace.matrix_RGB_to_XYZ,
            normalised_primary_matrix(
                colourspace.primaries, colourspace.whitepoint
            ),
            decimal=7,
        )

        np.testing.assert_array_almost_equal(
            colourspace.matrix_XYZ_to_RGB,
            np.linalg.inv(
                normalised_primary_matrix(
                    colourspace.primaries, colourspace.whitepoint
                )
            ),
            decimal=7,
        )

    def test_copy(self):
        """
        Test :meth:`colour.models.rgb.rgb_colourspace.RGB_Colourspace.copy`
        method.
        """

        self.assertIsNot(self._colourspace.copy(), self)


class TestXYZ_to_RGB(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.rgb_colourspace.XYZ_to_RGB` definition
    unit tests methods.
    """

    def test_XYZ_to_RGB(self):
        """Test :func:`colour.models.rgb.rgb_colourspace.XYZ_to_RGB` definition."""

        np.testing.assert_almost_equal(
            XYZ_to_RGB(
                np.array([0.21638819, 0.12570000, 0.03847493]),
                np.array([0.34570, 0.35850]),
                np.array([0.31270, 0.32900]),
                np.array(
                    [
                        [3.24062548, -1.53720797, -0.49862860],
                        [-0.96893071, 1.87575606, 0.04151752],
                        [0.05571012, -0.20402105, 1.05699594],
                    ]
                ),
                "Bradford",
                eotf_inverse_sRGB,
            ),
            np.array([0.70556599, 0.19109268, 0.22340812]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_RGB(
                np.array([0.21638819, 0.12570000, 0.03847493]),
                np.array([0.34570, 0.35850]),
                np.array([0.31270, 0.32900]),
                np.array(
                    [
                        [3.24062548, -1.53720797, -0.49862860],
                        [-0.96893071, 1.87575606, 0.04151752],
                        [0.05571012, -0.20402105, 1.05699594],
                    ]
                ),
                None,
                eotf_inverse_sRGB,
            ),
            np.array([0.72794579, 0.18180021, 0.17951580]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_RGB(
                np.array([0.21638819, 0.12570000, 0.03847493]),
                np.array([0.34570, 0.35850]),
                np.array([0.32168, 0.33767]),
                np.array(
                    [
                        [1.04981102, 0.00000000, -0.00009748],
                        [-0.49590302, 1.37331305, 0.09824004],
                        [0.00000000, 0.00000000, 0.99125202],
                    ]
                ),
            ),
            np.array([0.21959099, 0.06985815, 0.04703704]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_RGB(
                np.array([0.21638819, 0.12570000, 0.03847493]),
                np.array([0.34570, 0.35850]),
                np.array([0.31270, 0.32900, 1.00000]),
                np.array(
                    [
                        [3.24062548, -1.53720797, -0.49862860],
                        [-0.96893071, 1.87575606, 0.04151752],
                        [0.05571012, -0.20402105, 1.05699594],
                    ]
                ),
            ),
            np.array([0.45620801, 0.03079991, 0.04091883]),
            decimal=7,
        )

    def test_n_dimensional_XYZ_to_RGB(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.XYZ_to_RGB` definition
        n-dimensional support.
        """

        XYZ = np.array([0.21638819, 0.12570000, 0.03847493])
        W_R = np.array([0.34570, 0.35850])
        W_T = np.array([0.31270, 0.32900])
        M = np.array(
            [
                [3.24062548, -1.53720797, -0.49862860],
                [-0.96893071, 1.87575606, 0.04151752],
                [0.05571012, -0.20402105, 1.05699594],
            ]
        )
        RGB = XYZ_to_RGB(XYZ, W_R, W_T, M, "Bradford", eotf_inverse_sRGB)

        XYZ = np.tile(XYZ, (6, 1))
        RGB = np.tile(RGB, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_RGB(XYZ, W_R, W_T, M, "Bradford", eotf_inverse_sRGB),
            RGB,
            decimal=7,
        )

        W_R = np.tile(W_R, (6, 1))
        W_T = np.tile(W_T, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_RGB(XYZ, W_R, W_T, M, "Bradford", eotf_inverse_sRGB),
            RGB,
            decimal=7,
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        W_R = np.reshape(W_R, (2, 3, 2))
        W_T = np.reshape(W_T, (2, 3, 2))
        RGB = np.reshape(RGB, (2, 3, 3))
        np.testing.assert_almost_equal(
            XYZ_to_RGB(XYZ, W_R, W_T, M, "Bradford", eotf_inverse_sRGB),
            RGB,
            decimal=7,
        )

    def test_domain_range_scale_XYZ_to_RGB(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.XYZ_to_RGB` definition
        domain and range scale support.
        """

        XYZ = np.array([0.21638819, 0.12570000, 0.03847493])
        W_R = np.array([0.34570, 0.35850])
        W_T = np.array([0.31270, 0.32900])
        M = np.array(
            [
                [3.24062548, -1.53720797, -0.49862860],
                [-0.96893071, 1.87575606, 0.04151752],
                [0.05571012, -0.20402105, 1.05699594],
            ]
        )
        RGB = XYZ_to_RGB(XYZ, W_R, W_T, M)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_RGB(XYZ * factor, W_R, W_T, M),
                    RGB * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_RGB(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.XYZ_to_RGB` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            W_R = np.array(case[0:2])
            W_T = np.array(case[0:2])
            M = np.vstack([case, case, case]).reshape([3, 3])
            XYZ_to_RGB(XYZ, W_R, W_T, M)


class TestRGB_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.rgb_colourspace.RGB_to_XYZ` definition
    unit tests methods.
    """

    def test_RGB_to_XYZ(self):
        """Test :func:`colour.models.rgb.rgb_colourspace.RGB_to_XYZ` definition."""

        np.testing.assert_almost_equal(
            RGB_to_XYZ(
                np.array([0.70556599, 0.19109268, 0.22340812]),
                np.array([0.31270, 0.32900]),
                np.array([0.34570, 0.35850]),
                np.array(
                    [
                        [0.41240000, 0.35760000, 0.18050000],
                        [0.21260000, 0.71520000, 0.07220000],
                        [0.01930000, 0.11920000, 0.95050000],
                    ]
                ),
                "Bradford",
                eotf_sRGB,
            ),
            np.array([0.21638819, 0.12570000, 0.03847493]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_XYZ(
                np.array([0.72794579, 0.18180021, 0.17951580]),
                np.array([0.31270, 0.32900]),
                np.array([0.34570, 0.35850]),
                np.array(
                    [
                        [0.41240000, 0.35760000, 0.18050000],
                        [0.21260000, 0.71520000, 0.07220000],
                        [0.01930000, 0.11920000, 0.95050000],
                    ]
                ),
                None,
                eotf_sRGB,
            ),
            np.array([0.21638819, 0.12570000, 0.03847493]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_XYZ(
                np.array([0.21959099, 0.06985815, 0.04703704]),
                np.array([0.32168, 0.33767]),
                np.array([0.34570, 0.35850]),
                np.array(
                    [
                        [0.95255240, 0.00000000, 0.00009368],
                        [0.34396645, 0.72816610, -0.07213255],
                        [0.00000000, 0.00000000, 1.00882518],
                    ]
                ),
            ),
            np.array([0.21638819, 0.12570000, 0.03847493]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_XYZ(
                np.array([0.45620801, 0.03079991, 0.04091883]),
                np.array([0.31270, 0.32900, 1.00000]),
                np.array([0.34570, 0.35850]),
                np.array(
                    [
                        [0.41240000, 0.35760000, 0.18050000],
                        [0.21260000, 0.71520000, 0.07220000],
                        [0.01930000, 0.11920000, 0.95050000],
                    ]
                ),
            ),
            np.array([0.21638819, 0.12570000, 0.03847493]),
            decimal=7,
        )

    def test_n_dimensional_RGB_to_XYZ(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.RGB_to_XYZ` definition
        n-dimensional support.
        """

        RGB = np.array([0.70556599, 0.19109268, 0.22340812])
        W_R = np.array([0.31270, 0.32900])
        W_T = np.array([0.34570, 0.35850])
        M = np.array(
            [
                [0.41240000, 0.35760000, 0.18050000],
                [0.21260000, 0.71520000, 0.07220000],
                [0.01930000, 0.11920000, 0.95050000],
            ]
        )
        XYZ = RGB_to_XYZ(RGB, W_R, W_T, M, "Bradford", eotf_sRGB)

        RGB = np.tile(RGB, (6, 1))
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_almost_equal(
            RGB_to_XYZ(RGB, W_R, W_T, M, "Bradford", eotf_sRGB), XYZ, decimal=7
        )

        W_R = np.tile(W_R, (6, 1))
        W_T = np.tile(W_T, (6, 1))
        np.testing.assert_almost_equal(
            RGB_to_XYZ(RGB, W_R, W_T, M, "Bradford", eotf_sRGB), XYZ, decimal=7
        )

        RGB = np.reshape(RGB, (2, 3, 3))
        W_R = np.reshape(W_R, (2, 3, 2))
        W_T = np.reshape(W_T, (2, 3, 2))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_almost_equal(
            RGB_to_XYZ(RGB, W_R, W_T, M, "Bradford", eotf_sRGB), XYZ, decimal=7
        )

    def test_domain_range_scale_XYZ_to_RGB(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.RGB_to_XYZ` definition
        domain and range scale support.
        """

        RGB = np.array([0.45620801, 0.03079991, 0.04091883])
        W_R = np.array([0.31270, 0.32900])
        W_T = np.array([0.34570, 0.35850])
        M = np.array(
            [
                [3.24062548, -1.53720797, -0.49862860],
                [-0.96893071, 1.87575606, 0.04151752],
                [0.05571012, -0.20402105, 1.05699594],
            ]
        )
        XYZ = RGB_to_XYZ(RGB, W_R, W_T, M)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    RGB_to_XYZ(RGB * factor, W_R, W_T, M),
                    XYZ * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_RGB_to_XYZ(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.RGB_to_XYZ` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            RGB = np.array(case)
            W_R = np.array(case[0:2])
            W_T = np.array(case[0:2])
            M = np.vstack([case, case, case]).reshape([3, 3])
            RGB_to_XYZ(RGB, W_R, W_T, M)


class TestMatrix_RGB_to_RGB(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.rgb_colourspace.matrix_RGB_to_RGB`
    definition unit tests methods.
    """

    def test_matrix_RGB_to_RGB(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.matrix_RGB_to_RGB`
        definition.
        """

        aces_2065_1_colourspace = RGB_COLOURSPACES["ACES2065-1"]
        aces_cg_colourspace = RGB_COLOURSPACES["ACEScg"]
        sRGB_colourspace = RGB_COLOURSPACES["sRGB"]

        np.testing.assert_almost_equal(
            matrix_RGB_to_RGB(aces_2065_1_colourspace, sRGB_colourspace),
            np.array(
                [
                    [2.52164943, -1.13688855, -0.38491759],
                    [-0.27521355, 1.36970515, -0.09439245],
                    [-0.01592501, -0.14780637, 1.16380582],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_RGB_to_RGB(sRGB_colourspace, aces_2065_1_colourspace),
            np.array(
                [
                    [0.43958564, 0.38392940, 0.17653274],
                    [0.08953957, 0.81474984, 0.09568361],
                    [0.01738718, 0.10873911, 0.87382059],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_RGB_to_RGB(
                aces_2065_1_colourspace, aces_cg_colourspace, "Bradford"
            ),
            np.array(
                [
                    [1.45143932, -0.23651075, -0.21492857],
                    [-0.07655377, 1.17622970, -0.09967593],
                    [0.00831615, -0.00603245, 0.99771630],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_RGB_to_RGB(
                aces_2065_1_colourspace, sRGB_colourspace, "Bradford"
            ),
            np.array(
                [
                    [2.52140089, -1.13399575, -0.38756186],
                    [-0.27621406, 1.37259557, -0.09628236],
                    [-0.01532020, -0.15299256, 1.16838720],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_RGB_to_RGB(aces_2065_1_colourspace, sRGB_colourspace, None),
            np.array(
                [
                    [2.55809607, -1.11933692, -0.39181451],
                    [-0.27771575, 1.36589396, -0.09353075],
                    [-0.01711199, -0.14854588, 1.08104848],
                ]
            ),
            decimal=7,
        )


class TestRGB_to_RGB(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.rgb_colourspace.RGB_to_RGB` definition
    unit tests methods.
    """

    def test_RGB_to_RGB(self):
        """Test :func:`colour.models.rgb.rgb_colourspace.RGB_to_RGB` definition."""

        aces_2065_1_colourspace = RGB_COLOURSPACES["ACES2065-1"]
        sRGB_colourspace = RGB_COLOURSPACES["sRGB"]

        np.testing.assert_almost_equal(
            RGB_to_RGB(
                np.array([0.21931722, 0.06950287, 0.04694832]),
                aces_2065_1_colourspace,
                sRGB_colourspace,
            ),
            np.array([0.45595289, 0.03040780, 0.04087313]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_RGB(
                np.array([0.45595571, 0.03039702, 0.04087245]),
                sRGB_colourspace,
                aces_2065_1_colourspace,
            ),
            np.array([0.21931722, 0.06950287, 0.04694832]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_RGB(
                np.array([0.21931722, 0.06950287, 0.04694832]),
                aces_2065_1_colourspace,
                sRGB_colourspace,
                "Bradford",
            ),
            np.array([0.45597530, 0.03030054, 0.04086041]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_RGB(
                np.array([0.21931722, 0.06950287, 0.04694832]),
                aces_2065_1_colourspace,
                sRGB_colourspace,
                None,
            ),
            np.array([0.46484236, 0.02963459, 0.03667609]),
            decimal=7,
        )

        aces_cg_colourspace = RGB_COLOURSPACES["ACEScg"]
        aces_cc_colourspace = RGB_COLOURSPACES["ACEScc"]

        np.testing.assert_almost_equal(
            RGB_to_RGB(
                np.array([0.21931722, 0.06950287, 0.04694832]),
                aces_cg_colourspace,
                aces_cc_colourspace,
                apply_cctf_decoding=True,
                apply_cctf_encoding=True,
            ),
            np.array([0.42985679, 0.33522924, 0.30292336]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_RGB(
                np.array([0.46956438, 0.48137533, 0.43788601]),
                aces_cc_colourspace,
                sRGB_colourspace,
                apply_cctf_decoding=True,
                apply_cctf_encoding=True,
            ),
            np.array([0.60983062, 0.67896356, 0.50435764]),
            decimal=7,
        )

        np.testing.assert_equal(
            RGB_to_RGB(
                np.array([0.21931722, 0.06950287, 0.04694832]),
                aces_2065_1_colourspace,
                RGB_COLOURSPACES["ProPhoto RGB"],
                apply_cctf_encoding=True,
                out_int=True,
            ),
            np.array([120, 59, 46]),
        )

    def test_n_dimensional_RGB_to_RGB(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.RGB_to_RGB` definition
        n-dimensional support.
        """

        aces_2065_1_colourspace = RGB_COLOURSPACES["ACES2065-1"]
        sRGB_colourspace = RGB_COLOURSPACES["sRGB"]
        RGB_i = np.array([0.21931722, 0.06950287, 0.04694832])
        RGB_o = RGB_to_RGB(RGB_i, aces_2065_1_colourspace, sRGB_colourspace)

        RGB_i = np.tile(RGB_i, (6, 1))
        RGB_o = np.tile(RGB_o, (6, 1))
        np.testing.assert_almost_equal(
            RGB_to_RGB(RGB_i, aces_2065_1_colourspace, sRGB_colourspace),
            RGB_o,
            decimal=7,
        )

        RGB_i = np.reshape(RGB_i, (2, 3, 3))
        RGB_o = np.reshape(RGB_o, (2, 3, 3))
        np.testing.assert_almost_equal(
            RGB_to_RGB(RGB_i, aces_2065_1_colourspace, sRGB_colourspace),
            RGB_o,
            decimal=7,
        )

    def test_domain_range_scale_XYZ_to_RGB(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.RGB_to_RGB` definition
        domain and range scale support.
        """

        aces_2065_1_colourspace = RGB_COLOURSPACES["ACES2065-1"]
        sRGB_colourspace = RGB_COLOURSPACES["sRGB"]
        RGB_i = np.array([0.21931722, 0.06950287, 0.04694832])
        RGB_o = RGB_to_RGB(RGB_i, aces_2065_1_colourspace, sRGB_colourspace)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    RGB_to_RGB(
                        RGB_i * factor,
                        aces_2065_1_colourspace,
                        sRGB_colourspace,
                    ),
                    RGB_o * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_RGB_to_RGB(self):
        """
        Test :func:`colour.models.rgb.rgb_colourspace.RGB_to_RGB` definition
        nan support.
        """

        aces_2065_1_colourspace = RGB_COLOURSPACES["ACES2065-1"]
        sRGB_colourspace = RGB_COLOURSPACES["sRGB"]

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            RGB = np.array(case)
            RGB_to_RGB(RGB, aces_2065_1_colourspace, sRGB_colourspace)


if __name__ == "__main__":
    unittest.main()



"""
Display P3 Colourspace
======================

Defines the *Display P3* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_DISPLAY_P3`.

References
----------
-   :cite:`AppleInc.2019` : Apple. (2019). Apple Inc. (2019). displayP3.
    Retrieved December 18, 2019, from https://developer.apple.com/\
documentation/coregraphics/cgcolorspace/1408916-displayp3
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    eotf_inverse_sRGB,
    eotf_sRGB,
    normalised_primary_matrix,
)
from colour.models.rgb.datasets import RGB_COLOURSPACE_DCI_P3

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-science@googlegroups.com"
__status__ = "Production"

__all__ = [
    "PRIMARIES_DISPLAY_P3",
    "WHITEPOINT_NAME_DISPLAY_P3",
    "CCS_WHITEPOINT_DISPLAY_P3",
    "MATRIX_DISPLAY_P3_TO_XYZ",
    "MATRIX_XYZ_TO_DISPLAY_P3",
    "RGB_COLOURSPACE_DISPLAY_P3",
]

PRIMARIES_DISPLAY_P3: NDArray = RGB_COLOURSPACE_DCI_P3.primaries
"""*Display P3* colourspace primaries."""

WHITEPOINT_NAME_DISPLAY_P3: str = "D65"
"""*Display P3* colourspace whitepoint name."""

CCS_WHITEPOINT_DISPLAY_P3: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_DISPLAY_P3]
"""*Display P3* colourspace whitepoint chromaticity coordinates."""

MATRIX_DISPLAY_P3_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_DISPLAY_P3, CCS_WHITEPOINT_DISPLAY_P3
)
"""*Display P3* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_DISPLAY_P3: NDArray = np.linalg.inv(MATRIX_DISPLAY_P3_TO_XYZ)
"""*CIE XYZ* tristimulus values to *Display P3* colourspace matrix."""

RGB_COLOURSPACE_DISPLAY_P3: RGB_Colourspace = RGB_Colourspace(
    "Display P3",
    PRIMARIES_DISPLAY_P3,
    CCS_WHITEPOINT_DISPLAY_P3,
    WHITEPOINT_NAME_DISPLAY_P3,
    MATRIX_DISPLAY_P3_TO_XYZ,
    MATRIX_XYZ_TO_DISPLAY_P3,
    eotf_inverse_sRGB,
    eotf_sRGB,
)
RGB_COLOURSPACE_DISPLAY_P3.__doc__ = """
*Display P3* colourspace.

References
----------
:cite:`AppleInc.2019`
"""



"""Defines the unit tests for the :mod:`colour.algebra.geometry` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.algebra import (
    normalise_vector,
    euclidean_distance,
    manhattan_distance,
    extend_line_segment,
    intersect_line_segments,
    ellipse_coefficients_general_form,
    ellipse_coefficients_canonical_form,
    point_at_angle_on_ellipse,
    ellipse_fitting_Halir1998,
)
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestNormaliseVector",
    "TestEuclideanDistance",
    "TestManhattanDistance",
    "TestExtendLineSegment",
    "TestIntersectLineSegments",
    "TestEllipseCoefficientsCanonicalForm",
    "TestEllipseCoefficientsGeneralForm",
    "TestPointAtAngleOnEllipse",
    "TestEllipseFittingHalir1998",
]


class TestNormaliseVector(unittest.TestCase):
    """
    Define :func:`colour.algebra.geometry.normalise_vector` definition unit
    tests methods.
    """

    def test_normalise_vector(self):
        """Test :func:`colour.algebra.geometry.normalise_vector` definition."""

        np.testing.assert_almost_equal(
            normalise_vector(np.array([0.20654008, 0.12197225, 0.05136952])),
            np.array([0.84197033, 0.49722560, 0.20941026]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            normalise_vector(np.array([0.14222010, 0.23042768, 0.10495772])),
            np.array([0.48971705, 0.79344877, 0.36140872]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            normalise_vector(np.array([0.07818780, 0.06157201, 0.28099326])),
            np.array([0.26229003, 0.20655044, 0.94262445]),
            decimal=7,
        )


class TestEuclideanDistance(unittest.TestCase):
    """
    Define :func:`colour.algebra.geometry.euclidean_distance` definition unit
    tests methods.
    """

    def test_euclidean_distance(self):
        """Test :func:`colour.algebra.geometry.euclidean_distance` definition."""

        self.assertAlmostEqual(
            euclidean_distance(
                np.array([100.00000000, 21.57210357, 272.22819350]),
                np.array([100.00000000, 426.67945353, 72.39590835]),
            ),
            451.71330197,
            places=7,
        )

        self.assertAlmostEqual(
            euclidean_distance(
                np.array([100.00000000, 21.57210357, 272.22819350]),
                np.array([100.00000000, 74.05216981, 276.45318193]),
            ),
            52.64986116,
            places=7,
        )

        self.assertAlmostEqual(
            euclidean_distance(
                np.array([100.00000000, 21.57210357, 272.22819350]),
                np.array([100.00000000, 8.32281957, -73.58297716]),
            ),
            346.06489172,
            places=7,
        )

    def test_n_dimensional_euclidean_distance(self):
        """
        Test :func:`colour.algebra.geometry.euclidean_distance` definition
        n-dimensional arrays support.
        """

        a = np.array([100.00000000, 21.57210357, 272.22819350])
        b = np.array([100.00000000, 426.67945353, 72.39590835])
        distance = euclidean_distance(a, b)

        a = np.tile(a, (6, 1))
        b = np.tile(b, (6, 1))
        distance = np.tile(distance, 6)
        np.testing.assert_almost_equal(
            euclidean_distance(a, b), distance, decimal=7
        )

        a = np.reshape(a, (2, 3, 3))
        b = np.reshape(b, (2, 3, 3))
        distance = np.reshape(distance, (2, 3))
        np.testing.assert_almost_equal(
            euclidean_distance(a, b), distance, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_euclidean_distance(self):
        """
        Test :func:`colour.algebra.geometry.euclidean_distance` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            a = np.array(case)
            b = np.array(case)
            euclidean_distance(a, b)


class TestManhattanDistance(unittest.TestCase):
    """
    Define :func:`colour.algebra.geometry.manhattan_distance` definition unit
    tests methods.
    """

    def test_manhattan_distance(self):
        """Test :func:`colour.algebra.geometry.manhattan_distance` definition."""

        self.assertAlmostEqual(
            manhattan_distance(
                np.array([100.00000000, 21.57210357, 272.22819350]),
                np.array([100.00000000, 426.67945353, 72.39590835]),
            ),
            604.93963510999993,
            places=7,
        )

        self.assertAlmostEqual(
            manhattan_distance(
                np.array([100.00000000, 21.57210357, 272.22819350]),
                np.array([100.00000000, 74.05216981, 276.45318193]),
            ),
            56.705054670000052,
            places=7,
        )

        self.assertAlmostEqual(
            manhattan_distance(
                np.array([100.00000000, 21.57210357, 272.22819350]),
                np.array([100.00000000, 8.32281957, -73.58297716]),
            ),
            359.06045465999995,
            places=7,
        )

    def test_n_dimensional_manhattan_distance(self):
        """
        Test :func:`colour.algebra.geometry.manhattan_distance` definition
        n-dimensional arrays support.
        """

        a = np.array([100.00000000, 21.57210357, 272.22819350])
        b = np.array([100.00000000, 426.67945353, 72.39590835])
        distance = manhattan_distance(a, b)

        a = np.tile(a, (6, 1))
        b = np.tile(b, (6, 1))
        distance = np.tile(distance, 6)
        np.testing.assert_almost_equal(
            manhattan_distance(a, b), distance, decimal=7
        )

        a = np.reshape(a, (2, 3, 3))
        b = np.reshape(b, (2, 3, 3))
        distance = np.reshape(distance, (2, 3))
        np.testing.assert_almost_equal(
            manhattan_distance(a, b), distance, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_manhattan_distance(self):
        """
        Test :func:`colour.algebra.geometry.manhattan_distance` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            a = np.array(case)
            b = np.array(case)
            manhattan_distance(a, b)


class TestExtendLineSegment(unittest.TestCase):
    """
    Define :func:`colour.algebra.geometry.extend_line_segment` definition unit
    tests methods.
    """

    def test_extend_line_segment(self):
        """Test :func:`colour.algebra.geometry.extend_line_segment` definition."""

        np.testing.assert_almost_equal(
            extend_line_segment(
                np.array([0.95694934, 0.13720932]),
                np.array([0.28382835, 0.60608318]),
            ),
            np.array([-0.5367248, 1.17765341]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            extend_line_segment(
                np.array([0.95694934, 0.13720932]),
                np.array([0.28382835, 0.60608318]),
                5,
            ),
            np.array([-3.81893739, 3.46393435]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            extend_line_segment(
                np.array([0.95694934, 0.13720932]),
                np.array([0.28382835, 0.60608318]),
                -1,
            ),
            np.array([1.1043815, 0.03451295]),
            decimal=7,
        )


class TestIntersectLineSegments(unittest.TestCase):
    """
    Define :func:`colour.algebra.geometry.intersect_line_segments` definition
    unit tests methods.
    """

    def test_intersect_line_segments(self):
        """
        Test :func:`colour.algebra.geometry.intersect_line_segments`
        definition.
        """

        l_1 = np.array(
            [
                [[0.15416284, 0.7400497], [0.26331502, 0.53373939]],
                [[0.01457496, 0.91874701], [0.90071485, 0.03342143]],
            ]
        )
        l_2 = np.array(
            [
                [[0.95694934, 0.13720932], [0.28382835, 0.60608318]],
                [[0.94422514, 0.85273554], [0.00225923, 0.52122603]],
                [[0.55203763, 0.48537741], [0.76813415, 0.16071675]],
                [[0.01457496, 0.91874701], [0.90071485, 0.03342143]],
            ]
        )

        s = intersect_line_segments(l_1, l_2)

        np.testing.assert_almost_equal(
            s.xy,
            np.array(
                [
                    [
                        [np.nan, np.nan],
                        [0.22791841, 0.60064309],
                        [np.nan, np.nan],
                        [np.nan, np.nan],
                    ],
                    [
                        [0.42814517, 0.50555685],
                        [0.30560559, 0.62798382],
                        [0.7578749, 0.17613012],
                        [np.nan, np.nan],
                    ],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_array_equal(
            s.intersect,
            np.array([[False, True, False, False], [True, True, True, False]]),
        )

        np.testing.assert_array_equal(
            s.parallel,
            np.array(
                [[False, False, False, False], [False, False, False, True]]
            ),
        )

        np.testing.assert_array_equal(
            s.coincident,
            np.array(
                [[False, False, False, False], [False, False, False, True]]
            ),
        )


class TestEllipseCoefficientsCanonicalForm(unittest.TestCase):
    """
    Define :func:`colour.algebra.geometry.ellipse_coefficients_canonical_form`
    definition unit tests methods.
    """

    def test_ellipse_coefficients_canonical_form(self):
        """
        Test :func:`colour.algebra.geometry.\
ellipse_coefficients_canonical_form` definition.
        """

        np.testing.assert_almost_equal(
            ellipse_coefficients_canonical_form(
                np.array([2.5, -3.0, 2.5, -1.0, -1.0, -3.5])
            ),
            np.array([0.5, 0.5, 2, 1, 45]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            ellipse_coefficients_canonical_form(
                np.array([1.0, 0.0, 1.0, 0.0, 0.0, -1.0])
            ),
            np.array([0.0, 0.0, 1, 1, 0]),
            decimal=7,
        )


class TestEllipseCoefficientsGeneralForm(unittest.TestCase):
    """
    Define :func:`colour.algebra.geometry.ellipse_coefficients_general_form`
    definition unit tests methods.
    """

    def test_ellipse_coefficients_general_form(self):
        """
        Test :func:`colour.algebra.geometry.ellipse_coefficients_general_form`
        definition.
        """

        np.testing.assert_almost_equal(
            ellipse_coefficients_general_form(np.array([0.5, 0.5, 2, 1, 45])),
            np.array([2.5, -3.0, 2.5, -1.0, -1.0, -3.5]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            ellipse_coefficients_general_form(np.array([0.0, 0.0, 1, 1, 0])),
            np.array([1.0, 0.0, 1.0, 0.0, 0.0, -1.0]),
            decimal=7,
        )


class TestPointAtAngleOnEllipse(unittest.TestCase):
    """
    Define :func:`colour.algebra.geometry.point_at_angle_on_ellipse`
    definition unit tests methods.
    """

    def test_point_at_angle_on_ellipse(self):
        """
        Test :func:`colour.algebra.geometry.point_at_angle_on_ellipse`
        definition.
        """

        np.testing.assert_almost_equal(
            point_at_angle_on_ellipse(
                np.array([0, 90, 180, 270]), np.array([0.0, 0.0, 2, 1, 0])
            ),
            np.array([[2, 0], [0, 1], [-2, 0], [0, -1]]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            point_at_angle_on_ellipse(
                np.linspace(0, 360, 10), np.array([0.5, 0.5, 2, 1, 45])
            ),
            np.array(
                [
                    [1.91421356, 1.91421356],
                    [1.12883096, 2.03786992],
                    [0.04921137, 1.44193985],
                    [-0.81947922, 0.40526565],
                    [-1.07077081, -0.58708129],
                    [-0.58708129, -1.07077081],
                    [0.40526565, -0.81947922],
                    [1.44193985, 0.04921137],
                    [2.03786992, 1.12883096],
                    [1.91421356, 1.91421356],
                ]
            ),
            decimal=7,
        )


class TestEllipseFittingHalir1998(unittest.TestCase):
    """
    Define :func:`colour.algebra.geometry.ellipse_fitting_Halir1998`
    definition unit tests methods.
    """

    def test_ellipse_fitting_Halir1998(self):
        """
        Test :func:`colour.algebra.geometry.ellipse_fitting_Halir1998`
        definition.
        """

        np.testing.assert_almost_equal(
            ellipse_fitting_Halir1998(
                np.array([[2, 0], [0, 1], [-2, 0], [0, -1]])
            ),
            np.array(
                [
                    0.24253563,
                    0.00000000,
                    0.97014250,
                    0.00000000,
                    0.00000000,
                    -0.97014250,
                ]
            ),
            decimal=7,
        )


if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour.adaptation.cmccat2000."""

import numpy as np
import unittest
from itertools import permutations

from colour.adaptation.cmccat2000 import (
    chromatic_adaptation_forward_CMCCAT2000,
    chromatic_adaptation_inverse_CMCCAT2000,
)

from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestChromaticAdaptationForwardCMCCAT2000",
    "TestChromaticAdaptationInverseCMCCAT2000",
]


class TestChromaticAdaptationForwardCMCCAT2000(unittest.TestCase):
    """
    Define :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_forward_CMCCAT2000` definition unit tests methods.
    """

    def test_chromatic_adaptation_forward_CMCCAT2000(self):
        """
        Test :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_forward_CMCCAT2000` definition.
        """

        np.testing.assert_almost_equal(
            chromatic_adaptation_forward_CMCCAT2000(
                np.array([22.48, 22.74, 8.54]),
                np.array([111.15, 100.00, 35.20]),
                np.array([94.81, 100.00, 107.30]),
                200,
                200,
            ),
            np.array([19.52698326, 23.06833960, 24.97175229]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            chromatic_adaptation_forward_CMCCAT2000(
                np.array([0.14222010, 0.23042768, 0.10495772]) * 100,
                np.array([0.95045593, 1.00000000, 1.08905775]) * 100,
                np.array([1.09846607, 1.00000000, 0.35582280]) * 100,
                100,
                100,
            ),
            np.array([17.90511171, 22.75299363, 3.79837384]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            chromatic_adaptation_forward_CMCCAT2000(
                np.array([0.07818780, 0.06157201, 0.28099326]) * 100,
                np.array([0.95045593, 1.00000000, 1.08905775]) * 100,
                np.array([0.99144661, 1.00000000, 0.67315942]) * 100,
                100,
                100,
            ),
            np.array([6.76564344, 5.86585763, 18.40577315]),
            decimal=7,
        )

    def test_n_dimensional_chromatic_adaptation_forward_CMCCAT2000(self):
        """
        Test :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_forward_CMCCAT2000` definition n-dimensional arrays
        support.
        """

        XYZ = np.array([22.48, 22.74, 8.54])
        XYZ_w = np.array([111.15, 100.00, 35.20])
        XYZ_wr = np.array([94.81, 100.00, 107.30])
        L_A1 = 200
        L_A2 = 200
        XYZ_c = chromatic_adaptation_forward_CMCCAT2000(
            XYZ, XYZ_w, XYZ_wr, L_A1, L_A2
        )

        XYZ = np.tile(XYZ, (6, 1))
        XYZ_c = np.tile(XYZ_c, (6, 1))
        np.testing.assert_almost_equal(
            chromatic_adaptation_forward_CMCCAT2000(
                XYZ, XYZ_w, XYZ_wr, L_A1, L_A2
            ),
            XYZ_c,
            decimal=7,
        )

        XYZ_w = np.tile(XYZ_w, (6, 1))
        XYZ_wr = np.tile(XYZ_wr, (6, 1))
        L_A1 = np.tile(L_A1, 6)
        L_A2 = np.tile(L_A2, 6)
        np.testing.assert_almost_equal(
            chromatic_adaptation_forward_CMCCAT2000(
                XYZ, XYZ_w, XYZ_wr, L_A1, L_A2
            ),
            XYZ_c,
            decimal=7,
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        XYZ_w = np.reshape(XYZ_w, (2, 3, 3))
        XYZ_wr = np.reshape(XYZ_wr, (2, 3, 3))
        L_A1 = np.reshape(L_A1, (2, 3))
        L_A2 = np.reshape(L_A2, (2, 3))
        XYZ_c = np.reshape(XYZ_c, (2, 3, 3))
        np.testing.assert_almost_equal(
            chromatic_adaptation_forward_CMCCAT2000(
                XYZ, XYZ_w, XYZ_wr, L_A1, L_A2
            ),
            XYZ_c,
            decimal=7,
        )

    def test_domain_range_scale_chromatic_adaptation_CMCCAT2000(self):
        """
        Test :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_forward_CMCCAT2000` definition domain and range scale
        support.
        """

        XYZ = np.array([22.48, 22.74, 8.54])
        XYZ_w = np.array([111.15, 100.00, 35.20])
        XYZ_wr = np.array([94.81, 100.00, 107.30])
        L_A1 = 200
        L_A2 = 200
        XYZ_c = chromatic_adaptation_forward_CMCCAT2000(
            XYZ, XYZ_w, XYZ_wr, L_A1, L_A2
        )

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    chromatic_adaptation_forward_CMCCAT2000(
                        XYZ * factor,
                        XYZ_w * factor,
                        XYZ_wr * factor,
                        L_A1,
                        L_A2,
                    ),
                    XYZ_c * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_chromatic_adaptation_forward_CMCCAT2000(self):
        """
        Test :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_forward_CMCCAT2000` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            XYZ_w = np.array(case)
            XYZ_wr = np.array(case)
            L_A1 = case[0]
            L_A2 = case[0]
            chromatic_adaptation_forward_CMCCAT2000(
                XYZ, XYZ_w, XYZ_wr, L_A1, L_A2
            )


class TestChromaticAdaptationInverseCMCCAT2000(unittest.TestCase):
    """
    Define :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_inverse_CMCCAT2000` definition unit tests methods.
    """

    def test_chromatic_adaptation_inverse_CMCCAT2000(self):
        """
        Test :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_inverse_CMCCAT2000` definition.
        """

        np.testing.assert_almost_equal(
            chromatic_adaptation_inverse_CMCCAT2000(
                np.array([19.52698326, 23.06833960, 24.97175229]),
                np.array([111.15, 100.00, 35.20]),
                np.array([94.81, 100.00, 107.30]),
                200,
                200,
            ),
            np.array([22.48, 22.74, 8.54]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            chromatic_adaptation_inverse_CMCCAT2000(
                np.array([17.90511171, 22.75299363, 3.79837384]),
                np.array([0.95045593, 1.00000000, 1.08905775]) * 100,
                np.array([1.09846607, 1.00000000, 0.35582280]) * 100,
                100,
                100,
            ),
            np.array([0.14222010, 0.23042768, 0.10495772]) * 100,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            chromatic_adaptation_inverse_CMCCAT2000(
                np.array([6.76564344, 5.86585763, 18.40577315]),
                np.array([0.95045593, 1.00000000, 1.08905775]) * 100,
                np.array([0.99144661, 1.00000000, 0.67315942]) * 100,
                100,
                100,
            ),
            np.array([0.07818780, 0.06157201, 0.28099326]) * 100,
            decimal=7,
        )

    def test_n_dimensional_chromatic_adaptation_inverse_CMCCAT2000(self):
        """
        Test :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_inverse_CMCCAT2000` definition n-dimensional arrays
        support.
        """

        XYZ_c = np.array([19.52698326, 23.06833960, 24.97175229])
        XYZ_w = np.array([111.15, 100.00, 35.20])
        XYZ_wr = np.array([94.81, 100.00, 107.30])
        L_A1 = 200
        L_A2 = 200
        XYZ = chromatic_adaptation_inverse_CMCCAT2000(
            XYZ_c, XYZ_w, XYZ_wr, L_A1, L_A2
        )

        XYZ_c = np.tile(XYZ_c, (6, 1))
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_almost_equal(
            chromatic_adaptation_inverse_CMCCAT2000(
                XYZ_c, XYZ_w, XYZ_wr, L_A1, L_A2
            ),
            XYZ,
            decimal=7,
        )

        XYZ_w = np.tile(XYZ_w, (6, 1))
        XYZ_wr = np.tile(XYZ_wr, (6, 1))
        L_A1 = np.tile(L_A1, 6)
        L_A2 = np.tile(L_A2, 6)
        np.testing.assert_almost_equal(
            chromatic_adaptation_inverse_CMCCAT2000(
                XYZ_c, XYZ_w, XYZ_wr, L_A1, L_A2
            ),
            XYZ,
            decimal=7,
        )

        XYZ_c = np.reshape(XYZ_c, (2, 3, 3))
        XYZ_w = np.reshape(XYZ_w, (2, 3, 3))
        XYZ_wr = np.reshape(XYZ_wr, (2, 3, 3))
        L_A1 = np.reshape(L_A1, (2, 3))
        L_A2 = np.reshape(L_A2, (2, 3))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_almost_equal(
            chromatic_adaptation_inverse_CMCCAT2000(
                XYZ_c, XYZ_w, XYZ_wr, L_A1, L_A2
            ),
            XYZ,
            decimal=7,
        )

    def test_domain_range_scale_chromatic_adaptation_CMCCAT2000(self):
        """
        Test :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_inverse_CMCCAT2000` definition domain and range scale
        support.
        """

        XYZ_c = np.array([19.52698326, 23.06833960, 24.97175229])
        XYZ_w = np.array([111.15, 100.00, 35.20])
        XYZ_wr = np.array([94.81, 100.00, 107.30])
        L_A1 = 200
        L_A2 = 200
        XYZ = chromatic_adaptation_inverse_CMCCAT2000(
            XYZ_c, XYZ_w, XYZ_wr, L_A1, L_A2
        )

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    chromatic_adaptation_inverse_CMCCAT2000(
                        XYZ_c * factor,
                        XYZ_w * factor,
                        XYZ_wr * factor,
                        L_A1,
                        L_A2,
                    ),
                    XYZ * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_chromatic_adaptation_inverse_CMCCAT2000(self):
        """
        Test :func:`colour.adaptation.cmccat2000.\
chromatic_adaptation_inverse_CMCCAT2000` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ_c = np.array(case)
            XYZ_w = np.array(case)
            XYZ_wr = np.array(case)
            L_A1 = case[0]
            L_A2 = case[0]
            chromatic_adaptation_inverse_CMCCAT2000(
                XYZ_c, XYZ_w, XYZ_wr, L_A1, L_A2
            )


if __name__ == "__main__":
    unittest.main()



"""
Munsell Renotation System
=========================

Defines various objects for *Munsell Renotation System* computations:

-   :func:`colour.notation.munsell_value_Priest1920`: *Munsell* value :math:`V`
    computation of given *luminance* :math:`Y` using
    *Priest, Gibson and MacNicholas (1920)* method.
-   :func:`colour.notation.munsell_value_Munsell1933`: *Munsell* value
    :math:`V` computation of given *luminance* :math:`Y` using
    *Munsell, Sloan and Godlove (1933)* method.
-   :func:`colour.notation.munsell_value_Moon1943`: *Munsell* value :math:`V`
    computation of given *luminance* :math:`Y` using
    *Moon and Spencer (1943)* method.
-   :func:`colour.notation.munsell_value_Saunderson1944`: *Munsell* value
    :math:`V` computation of given *luminance* :math:`Y` using
    *Saunderson and Milner (1944)* method.
-   :func:`colour.notation.munsell_value_Ladd1955`: *Munsell* value :math:`V`
    computation of given *luminance* :math:`Y` using *Ladd and Pinney (1955)*
    method.
-   :func:`colour.notation.munsell_value_McCamy1987`: *Munsell* value :math:`V`
    computation of given *luminance* :math:`Y` using *McCamy (1987)* method.
-   :func:`colour.notation.munsell_value_ASTMD1535`: *Munsell* value
    :math:`V` computation of given *luminance* :math:`Y` using
    *ASTM D1535-08e1* method.
-   :attr:`colour.MUNSELL_VALUE_METHODS`: Supported *Munsell* value
    computation methods.
-   :func:`colour.munsell_value`: *Munsell* value :math:`V` computation of
    given *luminance* :math:`Y` using given method.
-   :func:`colour.munsell_colour_to_xyY`
-   :func:`colour.xyY_to_munsell_colour`

Notes
-----
-   The Munsell Renotation data commonly available within the *all.dat*,
    *experimental.dat* and *real.dat* files features *CIE xyY* colourspace values
    that are scaled by a :math:`1 / 0.975 \\simeq 1.02568` factor. If you are
    performing conversions using *Munsell* *Colorlab* specification,
    e.g. *2.5R 9/2*, according to *ASTM D1535-08e1* method, you should not
    scale the output :math:`Y` Luminance. However, if you use directly the
    *CIE xyY* colourspace values from the Munsell Renotation data data, you
    should scale the :math:`Y` Luminance before conversions by a :math:`0.975`
    factor.

    *ASTM D1535-08e1* states that::

        The coefficients of this equation are obtained from the 1943 equation
        by multiplying each coefficient by 0.975, the reflectance factor of
        magnesium oxide with respect to the perfect reflecting diffuser, and
        rounding to ve digits of precision.

References
----------
-   :cite:`ASTMInternational1989a` : ASTM International. (1989). ASTM D1535-89
    - Standard Practice for Specifying Color by the Munsell System (pp. 1-29).
    Retrieved September 25, 2014, from
    http://www.astm.org/DATABASE.CART/HISTORICAL/D1535-89.htm
-   :cite:`Centore2012a` : Centore, P. (2012). An open-source inversion
    algorithm for the Munsell renotation. Color Research & Application, 37(6),
    455-464. doi:10.1002/col.20715
-   :cite:`Centore2014k` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellRenotationRoutines/MunsellHueToASTMHue.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014l` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellSystemRoutines/LinearVsRadialInterpOnRenotationOvoid.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014m` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellRenotationRoutines/MunsellToxyY.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014n` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellRenotationRoutines/FindHueOnRenotationOvoid.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014o` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellSystemRoutines/BoundingRenotationHues.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014p` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellRenotationRoutines/xyYtoMunsell.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014q` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellRenotationRoutines/MunsellToxyForIntegerMunsellValue.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014r` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellRenotationRoutines/MaxChromaForExtrapolatedRenotation.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014s` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellRenotationRoutines/MunsellHueToChromDiagHueAngle.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014t` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    MunsellRenotationRoutines/ChromDiagHueAngleToMunsellHue.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centore2014u` : Centore, P. (2014).
    MunsellAndKubelkaMunkToolboxApr2014 -
    GeneralRoutines/CIELABtoApproxMunsellSpec.m.
    https://github.com/colour-science/MunsellAndKubelkaMunkToolbox
-   :cite:`Centorea` : Centore, P. (n.d.). The Munsell and Kubelka-Munk
    Toolbox. Retrieved January 23, 2018, from
    http://www.munsellcolourscienceforpainters.com/\
MunsellAndKubelkaMunkToolbox/MunsellAndKubelkaMunkToolbox.html
-   :cite:`Wikipedia2007c` : Nayatani, Y., Sobagaki, H., & Yano, K. H. T.
    (1995). Lightness dependency of chroma scales of a nonlinear
    color-appearance model and its latest formulation. Color Research &
    Application, 20(3), 156-167. doi:10.1002/col.5080200305
"""

from __future__ import annotations

import numpy as np
import re

from colour.algebra import (
    Extrapolator,
    LinearInterpolator,
    cartesian_to_cylindrical,
    euclidean_distance,
    polar_to_cartesian,
    spow,
)
from colour.colorimetry import CCS_ILLUMINANTS, luminance_ASTMD1535
from colour.constants import (
    INTEGER_THRESHOLD,
    FLOATING_POINT_NUMBER_PATTERN,
)
from colour.hints import (
    ArrayLike,
    Boolean,
    Dict,
    Floating,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Integer,
    Literal,
    NDArray,
    Optional,
    StrOrArrayLike,
    StrOrNDArray,
    Tuple,
    Union,
)
from colour.models import Lab_to_LCHab, XYZ_to_Lab, XYZ_to_xy, xyY_to_XYZ
from colour.volume import is_within_macadam_limits
from colour.notation import MUNSELL_COLOURS_ALL
from colour.utilities import (
    CACHE_REGISTRY,
    CaseInsensitiveMapping,
    Lookup,
    as_float,
    as_float_array,
    as_float_scalar,
    as_int_scalar,
    attest,
    domain_range_scale,
    from_range_1,
    from_range_10,
    get_domain_range_scale,
    to_domain_1,
    to_domain_10,
    to_domain_100,
    is_integer,
    is_numeric,
    tsplit,
    tstack,
    usage_warning,
    validate_method,
)

__author__ = "Colour Developers, Paul Centore"
__copyright__ = "Copyright 2013 Colour Developers"
__copyright__ += ", "
__copyright__ += (
    "The Munsell and Kubelka-Munk Toolbox: Copyright  2010-2018 Paul Centore "
    "(Gales Ferry, CT 06335, USA); used by permission."
)
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "MUNSELL_GRAY_PATTERN",
    "MUNSELL_COLOUR_PATTERN",
    "MUNSELL_GRAY_FORMAT",
    "MUNSELL_COLOUR_FORMAT",
    "MUNSELL_GRAY_EXTENDED_FORMAT",
    "MUNSELL_COLOUR_EXTENDED_FORMAT",
    "MUNSELL_HUE_LETTER_CODES",
    "ILLUMINANT_NAME_MUNSELL",
    "CCS_ILLUMINANT_MUNSELL",
    "munsell_value_Priest1920",
    "munsell_value_Munsell1933",
    "munsell_value_Moon1943",
    "munsell_value_Saunderson1944",
    "munsell_value_Ladd1955",
    "munsell_value_McCamy1987",
    "munsell_value_ASTMD1535",
    "MUNSELL_VALUE_METHODS",
    "munsell_value",
    "munsell_specification_to_xyY",
    "munsell_colour_to_xyY",
    "xyY_to_munsell_specification",
    "xyY_to_munsell_colour",
    "parse_munsell_colour",
    "is_grey_munsell_colour",
    "normalise_munsell_specification",
    "munsell_colour_to_munsell_specification",
    "munsell_specification_to_munsell_colour",
    "xyY_from_renotation",
    "is_specification_in_renotation",
    "bounding_hues_from_renotation",
    "hue_to_hue_angle",
    "hue_angle_to_hue",
    "hue_to_ASTM_hue",
    "interpolation_method_from_renotation_ovoid",
    "xy_from_renotation_ovoid",
    "LCHab_to_munsell_specification",
    "maximum_chroma_from_renotation",
    "munsell_specification_to_xy",
]

MUNSELL_GRAY_PATTERN: str = f"N(?P<value>{FLOATING_POINT_NUMBER_PATTERN})"
MUNSELL_COLOUR_PATTERN: str = (
    f"(?P<hue>{FLOATING_POINT_NUMBER_PATTERN})\\s*"
    f"(?P<letter>BG|GY|YR|RP|PB|B|G|Y|R|P)\\s*"
    f"(?P<value>{FLOATING_POINT_NUMBER_PATTERN})\\s*\\/\\s*"
    f"(?P<chroma>[-+]?{FLOATING_POINT_NUMBER_PATTERN})"
)

MUNSELL_GRAY_FORMAT: str = "N{0}"
MUNSELL_COLOUR_FORMAT: str = "{0} {1}/{2}"
MUNSELL_GRAY_EXTENDED_FORMAT: str = "N{0:.{1}f}"
MUNSELL_COLOUR_EXTENDED_FORMAT: str = "{0:.{1}f}{2} {3:.{4}f}/{5:.{6}f}"

MUNSELL_HUE_LETTER_CODES: Lookup = Lookup(
    {
        "BG": 2,
        "GY": 4,
        "YR": 6,
        "RP": 8,
        "PB": 10,
        "B": 1,
        "G": 3,
        "Y": 5,
        "R": 7,
        "P": 9,
    }
)

ILLUMINANT_NAME_MUNSELL: str = "C"
CCS_ILLUMINANT_MUNSELL: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][ILLUMINANT_NAME_MUNSELL]

_MUNSELL_SPECIFICATIONS_CACHE: Dict = CACHE_REGISTRY.register_cache(
    f"{__name__}._MUNSELL_SPECIFICATIONS_CACHE"
)
_MUNSELL_VALUE_ASTM_D1535_08_INTERPOLATOR_CACHE: Dict = (
    CACHE_REGISTRY.register_cache(
        f"{__name__}._MUNSELL_VALUE_ASTM_D1535_08_INTERPOLATOR_CACHE"
    )
)
_MUNSELL_MAXIMUM_CHROMAS_FROM_RENOTATION_CACHE: Dict = (
    CACHE_REGISTRY.register_cache(
        f"{__name__}._MUNSELL_MAXIMUM_CHROMAS_FROM_RENOTATION_CACHE"
    )
)


def _munsell_specifications() -> NDArray:
    """
    Return the *Munsell Renotation System* specifications and caches them if
    not existing.

    The *Munsell Renotation System* data is stored in
    :attr:`colour.notation.MUNSELL_COLOURS` attribute in a 2 columns form::

        (
            (('2.5GY', 0.2, 2.0), (0.713, 1.414, 0.237)),
            (('5GY', 0.2, 2.0), (0.449, 1.145, 0.237)),
            ...,
            (('7.5GY', 0.2, 2.0), (0.262, 0.837, 0.237)),
        )

    The first column is converted from *Munsell* colour to specification using
    :func:`colour.notation.munsell.munsell_colour_to_munsell_specification`
    definition:

    ('2.5GY', 0.2, 2.0) --> (2.5, 0.2, 2.0, 4)

    Returns
    -------
    :class:`numpy.ndarray`
        *Munsell Renotation System* specifications.
    """

    global _MUNSELL_SPECIFICATIONS_CACHE

    if "All" in _MUNSELL_SPECIFICATIONS_CACHE:
        return _MUNSELL_SPECIFICATIONS_CACHE["All"]

    munsell_specifications = np.array(
        [
            munsell_colour_to_munsell_specification(
                MUNSELL_COLOUR_FORMAT.format(*colour[0])
            )
            for colour in MUNSELL_COLOURS_ALL
        ]
    )

    _MUNSELL_SPECIFICATIONS_CACHE["All"] = munsell_specifications

    return munsell_specifications


def _munsell_value_ASTMD1535_interpolator() -> Extrapolator:
    """
    Return the *Munsell* value interpolator for *ASTM D1535-08e1* method and
    caches it if not existing.

    Returns
    -------
    :class:`colour.Extrapolator`
        *Munsell* value interpolator for *ASTM D1535-08e1* method.
    """

    global _MUNSELL_VALUE_ASTM_D1535_08_INTERPOLATOR_CACHE

    if "ASTM D1535-08 Interpolator" in (
        _MUNSELL_VALUE_ASTM_D1535_08_INTERPOLATOR_CACHE
    ):
        return _MUNSELL_VALUE_ASTM_D1535_08_INTERPOLATOR_CACHE[
            "ASTM D1535-08 Interpolator"
        ]

    munsell_values = np.arange(0, 10, 0.001)
    interpolator = LinearInterpolator(
        luminance_ASTMD1535(munsell_values), munsell_values
    )
    extrapolator = Extrapolator(interpolator)

    _MUNSELL_VALUE_ASTM_D1535_08_INTERPOLATOR_CACHE[
        "ASTM D1535-08 Interpolator"
    ] = extrapolator

    return extrapolator


def _munsell_maximum_chromas_from_renotation() -> Tuple[
    Tuple[Tuple[Floating, Floating, Floating], Floating], ...
]:
    """
    Return the maximum *Munsell* chromas from *Munsell Renotation System* data
    and caches them if not existing.

    Returns
    -------
    :class:`tuple`
        Maximum *Munsell* chromas.
    """

    global _MUNSELL_MAXIMUM_CHROMAS_FROM_RENOTATION_CACHE

    if "Maximum Chromas From Renotation" in (
        _MUNSELL_MAXIMUM_CHROMAS_FROM_RENOTATION_CACHE
    ):
        return _MUNSELL_MAXIMUM_CHROMAS_FROM_RENOTATION_CACHE[
            "Maximum Chromas From Renotation"
        ]

    chromas: Dict[Tuple[Floating, Floating, Floating], Floating] = {}
    for munsell_colour in MUNSELL_COLOURS_ALL:
        hue, value, chroma, code = tsplit(
            munsell_colour_to_munsell_specification(
                MUNSELL_COLOUR_FORMAT.format(*munsell_colour[0])
            )
        )
        index = (hue, value, code)
        if index in chromas:
            chroma = max([chromas[index], chroma])

        chromas[index] = chroma

    maximum_chromas_from_renotation = tuple(
        zip(chromas.keys(), chromas.values())
    )

    _MUNSELL_MAXIMUM_CHROMAS_FROM_RENOTATION_CACHE[
        "Maximum Chromas From Renotation"
    ] = maximum_chromas_from_renotation

    return maximum_chromas_from_renotation


def munsell_value_Priest1920(Y: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *Munsell* value :math:`V` of given *luminance* :math:`Y` using
    *Priest et al. (1920)* method.

    Parameters
    ----------
    Y
        *luminance* :math:`Y`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        *Munsell* value :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2007c`

    Examples
    --------
    >>> munsell_value_Priest1920(12.23634268)  # doctest: +ELLIPSIS
    3.4980484...
    """

    Y = to_domain_100(Y)

    V = 10 * np.sqrt(Y / 100)

    return as_float(from_range_10(V))


def munsell_value_Munsell1933(Y: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *Munsell* value :math:`V` of given *luminance* :math:`Y` using
    *Munsell et al. (1933)* method.

    Parameters
    ----------
    Y
        *luminance* :math:`Y`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        *Munsell* value :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2007c`

    Examples
    --------
    >>> munsell_value_Munsell1933(12.23634268)  # doctest: +ELLIPSIS
    4.1627702...
    """

    Y = to_domain_100(Y)

    V = np.sqrt(1.4742 * Y - 0.004743 * (Y * Y))

    return as_float(from_range_10(V))


def munsell_value_Moon1943(Y: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *Munsell* value :math:`V` of given *luminance* :math:`Y` using
    *Moon and Spencer (1943)* method.


    Parameters
    ----------
    Y
        *luminance* :math:`Y`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        *Munsell* value :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2007c`

    Examples
    --------
    >>> munsell_value_Moon1943(12.23634268)  # doctest: +ELLIPSIS
    4.0688120...
    """

    Y = to_domain_100(Y)

    V = 1.4 * spow(Y, 0.426)

    return as_float(from_range_10(V))


def munsell_value_Saunderson1944(Y: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *Munsell* value :math:`V` of given *luminance* :math:`Y` using
    *Saunderson and Milner (1944)* method.

    Parameters
    ----------
    Y
        *luminance* :math:`Y`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        *Munsell* value :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2007c`

    Examples
    --------
    >>> munsell_value_Saunderson1944(12.23634268)  # doctest: +ELLIPSIS
    4.0444736...
    """

    Y = to_domain_100(Y)

    V = 2.357 * spow(Y, 0.343) - 1.52

    return as_float(from_range_10(V))


def munsell_value_Ladd1955(Y: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *Munsell* value :math:`V` of given *luminance* :math:`Y` using
    *Ladd and Pinney (1955)* method.

    Parameters
    ----------
    Y
        *luminance* :math:`Y`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        *Munsell* value :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2007c`

    Examples
    --------
    >>> munsell_value_Ladd1955(12.23634268)  # doctest: +ELLIPSIS
    4.0511633...
    """

    Y = to_domain_100(Y)

    V = 2.468 * spow(Y, 1 / 3) - 1.636

    return as_float(from_range_10(V))


def munsell_value_McCamy1987(Y: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *Munsell* value :math:`V` of given *luminance* :math:`Y` using
    *McCamy (1987)* method.

    Parameters
    ----------
    Y
        *luminance* :math:`Y`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        *Munsell* value :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`ASTMInternational1989a`

    Examples
    --------
    >>> munsell_value_McCamy1987(12.23634268)  # doctest: +ELLIPSIS
    4.0814348...
    """

    Y = to_domain_100(Y)

    V = np.where(
        Y <= 0.9,
        0.87445 * spow(Y, 0.9967),
        2.49268 * spow(Y, 1 / 3)
        - 1.5614
        - (0.985 / (((0.1073 * Y - 3.084) ** 2) + 7.54))
        + (0.0133 / spow(Y, 2.3))
        + 0.0084 * np.sin(4.1 * spow(Y, 1 / 3) + 1)
        + (0.0221 / Y) * np.sin(0.39 * (Y - 2))
        - (0.0037 / (0.44 * Y)) * np.sin(1.28 * (Y - 0.53)),
    )

    return as_float(from_range_10(V))


def munsell_value_ASTMD1535(Y: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *Munsell* value :math:`V` of given *luminance* :math:`Y` using
    an inverse lookup table from *ASTM D1535-08e1* method.

    Parameters
    ----------
    Y
        *luminance* :math:`Y`

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        *Munsell* value :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    -   The *Munsell* value* computation with *ASTM D1535-08e1* method is only
        defined for domain [0, 100].

    References
    ----------
    :cite:`ASTMInternational1989a`

    Examples
    --------
    >>> munsell_value_ASTMD1535(12.23634268)  # doctest: +ELLIPSIS
    4.0824437...
    """

    Y = to_domain_100(Y)

    V = _munsell_value_ASTMD1535_interpolator()(Y)

    return as_float(from_range_10(V))


MUNSELL_VALUE_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Priest 1920": munsell_value_Priest1920,
        "Munsell 1933": munsell_value_Munsell1933,
        "Moon 1943": munsell_value_Moon1943,
        "Saunderson 1944": munsell_value_Saunderson1944,
        "Ladd 1955": munsell_value_Ladd1955,
        "McCamy 1987": munsell_value_McCamy1987,
        "ASTM D1535": munsell_value_ASTMD1535,
    }
)
MUNSELL_VALUE_METHODS.__doc__ = """
Supported *Munsell* value computation methods.

References
----------
:cite:`ASTMInternational1989a`, :cite:`Wikipedia2007c`

Aliases:

-   'astm2008': 'ASTM D1535'
"""
MUNSELL_VALUE_METHODS["astm2008"] = MUNSELL_VALUE_METHODS["ASTM D1535"]


def munsell_value(
    Y: FloatingOrArrayLike,
    method: Union[
        Literal[
            "ASTM D1535",
            "Ladd 1955",
            "McCamy 1987",
            "Moon 1943",
            "Munsell 1933",
            "Priest 1920",
            "Saunderson 1944",
        ],
        str,
    ] = "ASTM D1535",
) -> FloatingOrNDArray:
    """
    Return the *Munsell* value :math:`V` of given *luminance* :math:`Y` using
    given method.

    Parameters
    ----------
    Y
        *luminance* :math:`Y`.
    method
        Computation method.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        *Munsell* value :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`ASTMInternational1989a`, :cite:`Wikipedia2007c`

    Examples
    --------
    >>> munsell_value(12.23634268)  # doctest: +ELLIPSIS
    4.0824437...
    >>> munsell_value(12.23634268, method='Priest 1920') # doctest: +ELLIPSIS
    3.4980484...
    >>> munsell_value(12.23634268, method='Munsell 1933') # doctest: +ELLIPSIS
    4.1627702...
    >>> munsell_value(12.23634268, method='Moon 1943') # doctest: +ELLIPSIS
    4.0688120...
    >>> munsell_value(12.23634268, method='Saunderson 1944')
    ... # doctest: +ELLIPSIS
    4.0444736...
    >>> munsell_value(12.23634268, method='Ladd 1955') # doctest: +ELLIPSIS
    4.0511633...
    >>> munsell_value(12.23634268, method='McCamy 1987') # doctest: +ELLIPSIS
    4.0814348...
    """

    method = validate_method(method, MUNSELL_VALUE_METHODS)

    return MUNSELL_VALUE_METHODS[method](Y)


def _munsell_scale_factor() -> NDArray:
    """
    Return the domain-range scale factor for *Munsell Renotation System*.

    Returns
    -------
    :class:`numpy.ndarray`
        Domain-range scale factor for *Munsell Renotation System*.
    """

    return np.array([10, 10, 50 if get_domain_range_scale() == "1" else 2, 10])


def _munsell_specification_to_xyY(specification: ArrayLike) -> NDArray:
    """
    Convert given *Munsell* *Colorlab* specification to *CIE xyY* colourspace.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xyY* colourspace array.
    """

    specification = normalise_munsell_specification(specification)

    if is_grey_munsell_colour(specification):
        specification = as_float_array(to_domain_10(specification))
        hue, value, chroma, code = specification
    else:
        specification = to_domain_10(specification, _munsell_scale_factor())
        hue, value, chroma, code = specification
        code = as_int_scalar(code)

        attest(
            0 <= hue <= 10,
            f'"{specification}" specification hue must be normalised to '
            f"domain [0, 10]!",
        )

        attest(
            0 <= value <= 10,
            f'"{specification}" specification value must be normalised to '
            f"domain [0, 10]!",
        )

    with domain_range_scale("ignore"):
        Y = luminance_ASTMD1535(value)

    if is_integer(value):
        value_minus = value_plus = round(value)
    else:
        value_minus = np.floor(value)
        value_plus = value_minus + 1

    specification_minus = as_float_array(
        value_minus
        if is_grey_munsell_colour(specification)
        else [hue, value_minus, chroma, code]
    )
    x_minus, y_minus = tsplit(munsell_specification_to_xy(specification_minus))

    specification_plus = as_float_array(
        value_plus
        if (is_grey_munsell_colour(specification) or value_plus == 10)
        else [hue, value_plus, chroma, code]
    )
    x_plus, y_plus = tsplit(munsell_specification_to_xy(specification_plus))

    if value_minus == value_plus:
        x = x_minus
        y = y_minus
    else:
        with domain_range_scale("ignore"):
            Y_minus = as_float_array(luminance_ASTMD1535(value_minus))
            Y_plus = as_float_array(luminance_ASTMD1535(value_plus))

        Y_minus_plus = np.squeeze([Y_minus, Y_plus])
        x_minus_plus = np.squeeze([x_minus, x_plus])
        y_minus_plus = np.squeeze([y_minus, y_plus])

        x = LinearInterpolator(Y_minus_plus, x_minus_plus)(Y)
        y = LinearInterpolator(Y_minus_plus, y_minus_plus)(Y)

    return tstack([x, y, from_range_1(Y / 100)])


def munsell_specification_to_xyY(specification: ArrayLike) -> NDArray:
    """
    Convert given *Munsell* *Colorlab* specification to *CIE xyY* colourspace.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xyY* colourspace array.

    Notes
    -----
    +-------------------+-----------------------+---------------+
    | **Domain**        | **Scale - Reference** | **Scale - 1** |
    +===================+=======================+===============+
    | ``specification`` | ``hue``    : [0, 10]  | [0, 1]        |
    |                   |                       |               |
    |                   | ``value``  : [0, 10]  | [0, 1]        |
    |                   |                       |               |
    |                   | ``chroma`` : [0, 50]  | [0, 1]        |
    |                   |                       |               |
    |                   | ``code``   : [0, 10]  | [0, 1]        |
    +-------------------+-----------------------+---------------+

    +-------------------+-----------------------+---------------+
    | **Range**         | **Scale - Reference** | **Scale - 1** |
    +===================+=======================+===============+
    | ``xyY``           | [0, 1]                | [0, 1]        |
    +-------------------+-----------------------+---------------+

    References
    ----------
    :cite:`Centore2014m`

    Examples
    --------
    >>> munsell_specification_to_xyY(np.array([2.1, 8.0, 17.9, 4]))
    ... # doctest: +ELLIPSIS
    array([ 0.4400632...,  0.5522428...,  0.5761962...])
    >>> munsell_specification_to_xyY(np.array([np.nan, 8.9, np.nan, np.nan]))
    ... # doctest: +ELLIPSIS
    array([ 0.31006  ,  0.31616  ,  0.7461345...])
    """

    specification = as_float_array(specification)
    shape = list(specification.shape)

    xyY = [
        _munsell_specification_to_xyY(a)
        for a in specification.reshape([-1, 4])
    ]

    shape[-1] = 3

    return np.reshape(as_float_array(xyY), shape)


def munsell_colour_to_xyY(munsell_colour: StrOrArrayLike) -> NDArray:
    """
    Convert given *Munsell* colour to *CIE xyY* colourspace.

    Parameters
    ----------
    munsell_colour
        *Munsell* colour.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xyY* colourspace array.

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``xyY``   | [0, 1]                | [0, 1]        |
    +-----------+-----------------------+---------------+

    References
    ----------
    :cite:`Centorea`, :cite:`Centore2012a`

    Examples
    --------
    >>> munsell_colour_to_xyY('4.2YR 8.1/5.3')  # doctest: +ELLIPSIS
    array([ 0.3873694...,  0.3575165...,  0.59362   ])
    >>> munsell_colour_to_xyY('N8.9')  # doctest: +ELLIPSIS
    array([ 0.31006  ,  0.31616  ,  0.7461345...])
    """

    munsell_colour = np.array(munsell_colour)
    shape = list(munsell_colour.shape)

    specification = np.array(
        [
            munsell_colour_to_munsell_specification(a)
            for a in np.ravel(munsell_colour)
        ]
    )

    return munsell_specification_to_xyY(
        from_range_10(
            specification.reshape(shape + [4]), _munsell_scale_factor()
        )
    )


def _xyY_to_munsell_specification(xyY: ArrayLike) -> NDArray:
    """
    Convert from *CIE xyY* colourspace to *Munsell* *Colorlab* specification.

    Parameters
    ----------
    xyY
        *CIE xyY* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *Munsell* *Colorlab* specification.

    Raises
    ------
    ValueError
        If the given *CIE xyY* colourspace array is not within MacAdam
        limits.
    RuntimeError
        If the maximum iterations count has been reached without converging to
        a result.
    """

    x, y, Y = tsplit(xyY)
    Y = to_domain_1(Y)

    if not is_within_macadam_limits(xyY, ILLUMINANT_NAME_MUNSELL):
        usage_warning(
            f'"{xyY!r}" is not within "MacAdam" limits for illuminant '
            f'"{ILLUMINANT_NAME_MUNSELL}"!'
        )

    with domain_range_scale("ignore"):
        value = munsell_value_ASTMD1535(Y * 100)

    if is_integer(value):
        value = np.around(value)

    with domain_range_scale("ignore"):
        x_center, y_center, Y_center = tsplit(
            _munsell_specification_to_xyY(value)
        )

    rho_input, phi_input, _z_input = tsplit(
        cartesian_to_cylindrical([x - x_center, y - y_center, Y_center])
    )
    phi_input = np.degrees(phi_input)

    grey_threshold = 1e-7
    if rho_input < grey_threshold:
        return from_range_10(normalise_munsell_specification(value))

    X, Y, Z = xyY_to_XYZ([x, y, Y])
    xi, yi = CCS_ILLUMINANT_MUNSELL
    Xr, Yr, Zr = xyY_to_XYZ([xi, yi, Y])

    XYZ = np.array([X, Y, Z])
    XYZr = np.array([(1 / Yr) * Xr, 1, (1 / Yr) * Zr])

    Lab = XYZ_to_Lab(XYZ, XYZ_to_xy(XYZr))
    LCHab = Lab_to_LCHab(Lab)
    hue_initial, _value_initial, chroma_initial, code_initial = tsplit(
        LCHab_to_munsell_specification(LCHab)
    )
    specification_current = [
        hue_initial,
        value,
        (5 / 5.5) * chroma_initial,
        code_initial,
    ]

    convergence_threshold = 1e-7
    iterations_maximum = 64
    iterations = 0

    while iterations <= iterations_maximum:
        iterations += 1

        (
            hue_current,
            _value_current,
            chroma_current,
            code_current,
        ) = specification_current
        hue_angle_current = hue_to_hue_angle([hue_current, code_current])

        chroma_maximum = maximum_chroma_from_renotation(
            [hue_current, value, code_current]
        )
        if chroma_current > chroma_maximum:
            chroma_current = specification_current[2] = chroma_maximum

        with domain_range_scale("ignore"):
            x_current, y_current, _Y_current = tsplit(
                _munsell_specification_to_xyY(specification_current)
            )

        rho_current, phi_current, _z_current = tsplit(
            cartesian_to_cylindrical(
                [x_current - x_center, y_current - y_center, Y_center]
            )
        )
        phi_current = np.degrees(phi_current)
        phi_current_difference = (360 - phi_input + phi_current) % 360
        if phi_current_difference > 180:
            phi_current_difference -= 360

        phi_differences_data = [phi_current_difference]
        hue_angles_differences_data = [0]
        hue_angles = [hue_angle_current]

        iterations_maximum_inner = 16
        iterations_inner = 0
        extrapolate = False

        while (
            np.sign(np.min(phi_differences_data))
            == np.sign(np.max(phi_differences_data))
            and extrapolate is False
        ):
            iterations_inner += 1

            if iterations_inner > iterations_maximum_inner:
                # NOTE: This exception is likely never raised in practice:
                # 300K iterations with random numbers never reached this code
                # path, it is kept for consistency with the reference
                # implementation.
                raise RuntimeError(  # pragma: no cover
                    "Maximum inner iterations count reached without "
                    "convergence!"
                )

            hue_angle_inner = (
                hue_angle_current
                + iterations_inner * (phi_input - phi_current)
            ) % 360
            hue_angle_difference_inner = (
                iterations_inner * (phi_input - phi_current) % 360
            )
            if hue_angle_difference_inner > 180:
                hue_angle_difference_inner -= 360

            hue_inner, code_inner = hue_angle_to_hue(hue_angle_inner)

            with domain_range_scale("ignore"):
                x_inner, y_inner, _Y_inner = _munsell_specification_to_xyY(
                    [hue_inner, value, chroma_current, code_inner]
                )

            if len(phi_differences_data) >= 2:
                extrapolate = True

            if extrapolate is False:
                rho_inner, phi_inner, _z_inner = cartesian_to_cylindrical(
                    [x_inner - x_center, y_inner - y_center, Y_center]
                )
                phi_inner = np.degrees(phi_inner)
                phi_inner_difference = (360 - phi_input + phi_inner) % 360
                if phi_inner_difference > 180:
                    phi_inner_difference -= 360

                phi_differences_data.append(phi_inner_difference)
                hue_angles.append(hue_angle_inner)
                hue_angles_differences_data.append(hue_angle_difference_inner)

        phi_differences = np.array(phi_differences_data)
        hue_angles_differences = np.array(hue_angles_differences_data)

        phi_differences_indexes = phi_differences.argsort()

        phi_differences = phi_differences[phi_differences_indexes]
        hue_angles_differences = hue_angles_differences[
            phi_differences_indexes
        ]

        hue_angle_difference_new = (
            Extrapolator(
                LinearInterpolator(phi_differences, hue_angles_differences)
            )(0)
            % 360
        )
        hue_angle_new = (hue_angle_current + hue_angle_difference_new) % 360

        hue_new, code_new = hue_angle_to_hue(as_float_scalar(hue_angle_new))
        specification_current = [hue_new, value, chroma_current, code_new]

        with domain_range_scale("ignore"):
            x_current, y_current, _Y_current = _munsell_specification_to_xyY(
                specification_current
            )

        chroma_scale = 50 if get_domain_range_scale() == "1" else 2

        difference = euclidean_distance([x, y], [x_current, y_current])
        if difference < convergence_threshold:
            return from_range_10(
                np.array(specification_current),
                np.array([10, 10, chroma_scale, 10]),
            )

        # TODO: Consider refactoring implementation.
        (
            hue_current,
            _value_current,
            chroma_current,
            code_current,
        ) = specification_current
        chroma_maximum = maximum_chroma_from_renotation(
            [hue_current, value, code_current]
        )

        # NOTE: This condition is likely never "True" while producing a valid
        # "Munsell Specification" in practice: 100K iterations with random
        # numbers never reached this code path while producing a valid
        # "Munsell Specification".
        if chroma_current > chroma_maximum:
            chroma_current = specification_current[2] = chroma_maximum

        with domain_range_scale("ignore"):
            x_current, y_current, _Y_current = _munsell_specification_to_xyY(
                specification_current
            )

        rho_current, phi_current, _z_current = cartesian_to_cylindrical(
            [x_current - x_center, y_current - y_center, Y_center]
        )

        rho_bounds_data = [rho_current]
        chroma_bounds_data = [chroma_current]

        iterations_maximum_inner = 16
        iterations_inner = 0
        while not (
            np.min(rho_bounds_data) < rho_input < np.max(rho_bounds_data)
        ):
            iterations_inner += 1

            if iterations_inner > iterations_maximum_inner:
                raise RuntimeError(
                    "Maximum inner iterations count reached "
                    "without convergence!"
                )

            chroma_inner = (
                (rho_input / rho_current) ** iterations_inner
            ) * chroma_current
            if chroma_inner > chroma_maximum:
                chroma_inner = specification_current[2] = chroma_maximum

            specification_inner = (
                hue_current,
                value,
                chroma_inner,
                code_current,
            )

            with domain_range_scale("ignore"):
                x_inner, y_inner, _Y_inner = _munsell_specification_to_xyY(
                    specification_inner
                )

            rho_inner, phi_inner, _z_inner = cartesian_to_cylindrical(
                [x_inner - x_center, y_inner - y_center, Y_center]
            )

            rho_bounds_data.append(rho_inner)
            chroma_bounds_data.append(chroma_inner)

        rho_bounds = np.array(rho_bounds_data)
        chroma_bounds = np.array(chroma_bounds_data)

        rhos_bounds_indexes = rho_bounds.argsort()

        rho_bounds = rho_bounds[rhos_bounds_indexes]
        chroma_bounds = chroma_bounds[rhos_bounds_indexes]
        chroma_new = LinearInterpolator(rho_bounds, chroma_bounds)(rho_input)

        specification_current = [hue_current, value, chroma_new, code_current]

        with domain_range_scale("ignore"):
            x_current, y_current, _Y_current = _munsell_specification_to_xyY(
                specification_current
            )

        difference = euclidean_distance([x, y], [x_current, y_current])
        if difference < convergence_threshold:
            return from_range_10(
                np.array(specification_current),
                np.array([10, 10, chroma_scale, 10]),
            )

    # NOTE: This exception is likely never raised in practice: 300K iterations
    # with random numbers never reached this code path, it is kept for
    # consistency with the reference # implementation
    raise RuntimeError(  # pragma: no cover
        "Maximum outside iterations count reached without convergence!"
    )


def xyY_to_munsell_specification(xyY: ArrayLike) -> NDArray:
    """
    Convert from *CIE xyY* colourspace to *Munsell* *Colorlab* specification.

    Parameters
    ----------
    xyY
        *CIE xyY* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *Munsell* *Colorlab* specification.

    Raises
    ------
    ValueError
        If the given *CIE xyY* colourspace array is not within MacAdam
        limits.
    RuntimeError
        If the maximum iterations count has been reached without converging to
        a result.

    Notes
    -----
    +-------------------+-----------------------+---------------+
    | **Domain**        | **Scale - Reference** | **Scale - 1** |
    +===================+=======================+===============+
    | ``xyY``           | [0, 1]                | [0, 1]        |
    +-------------------+-----------------------+---------------+

    +-------------------+-----------------------+---------------+
    | **Range**         | **Scale - Reference** | **Scale - 1** |
    +===================+=======================+===============+
    | ``specification`` | ``hue``    : [0, 10]  | [0, 1]        |
    |                   |                       |               |
    |                   | ``value``  : [0, 10]  | [0, 1]        |
    |                   |                       |               |
    |                   | ``chroma`` : [0, 50]  | [0, 1]        |
    |                   |                       |               |
    |                   | ``code``   : [0, 10]  | [0, 1]        |
    +-------------------+-----------------------+---------------+

    References
    ----------
    :cite:`Centore2014p`

    Examples
    --------
    >>> xyY = np.array([0.38736945, 0.35751656, 0.59362000])
    >>> xyY_to_munsell_specification(xyY)  # doctest: +ELLIPSIS
    array([ 4.2000019...,  8.0999999...,  5.2999996...,  6.        ])
    """

    xyY = as_float_array(xyY)
    shape = list(xyY.shape)

    specification = [
        _xyY_to_munsell_specification(a) for a in xyY.reshape([-1, 3])
    ]

    shape[-1] = 4

    return np.reshape(as_float_array(specification), shape)


def xyY_to_munsell_colour(
    xyY: ArrayLike,
    hue_decimals: Integer = 1,
    value_decimals: Integer = 1,
    chroma_decimals: Integer = 1,
) -> StrOrNDArray:
    """
    Convert from *CIE xyY* colourspace to *Munsell* colour.

    Parameters
    ----------
    xyY
        *CIE xyY* colourspace array.
    hue_decimals
        Hue formatting decimals.
    value_decimals
        Value formatting decimals.
    chroma_decimals
        Chroma formatting decimals.

    Returns
    -------
    :class:`str` or :class:`numpy.ndarray`
        *Munsell* colour.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``xyY``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Centorea`, :cite:`Centore2012a`

    Examples
    --------
    >>> xyY = np.array([0.38736945, 0.35751656, 0.59362000])
    >>> xyY_to_munsell_colour(xyY)
    '4.2YR 8.1/5.3'
    """

    specification = to_domain_10(
        xyY_to_munsell_specification(xyY), _munsell_scale_factor()
    )
    shape = list(specification.shape)
    decimals = (hue_decimals, value_decimals, chroma_decimals)

    munsell_colour = np.reshape(
        np.array(
            [
                munsell_specification_to_munsell_colour(a, *decimals)
                for a in specification.reshape([-1, 4])
            ]
        ),
        shape[:-1],
    )

    return str(munsell_colour) if shape == [4] else munsell_colour


def parse_munsell_colour(munsell_colour: str) -> NDArray:
    """
    Parse given *Munsell* colour and returns an intermediate *Munsell*
    *Colorlab* specification.

    Parameters
    ----------
    munsell_colour
        *Munsell* colour.

    Returns
    -------
    :class:`numpy.ndarray`
        Intermediate *Munsell* *Colorlab* specification.

    Raises
    ------
    ValueError
        If the given specification is not a valid *Munsell Renotation System*
        colour specification.

    Examples
    --------
    >>> parse_munsell_colour('N5.2')
    array([ nan,  5.2,  nan,  nan])
    >>> parse_munsell_colour('0YR 2.0/4.0')
    array([ 0.,  2.,  4.,  6.])
    """

    match = re.match(MUNSELL_GRAY_PATTERN, munsell_colour, flags=re.IGNORECASE)
    if match:
        return tstack(
            [
                np.nan,
                match.group("value"),
                np.nan,
                np.nan,
            ]
        )

    match = re.match(
        MUNSELL_COLOUR_PATTERN, munsell_colour, flags=re.IGNORECASE
    )
    if match:
        return tstack(
            [
                match.group("hue"),
                match.group("value"),
                match.group("chroma"),
                MUNSELL_HUE_LETTER_CODES[match.group("letter").upper()],
            ]
        )

    raise ValueError(
        f'"{munsell_colour}" is not a valid "Munsell Renotation System" '
        f"colour specification!"
    )


def is_grey_munsell_colour(specification: ArrayLike) -> Boolean:
    """
    Return if given *Munsell* *Colorlab* specification is a grey colour.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.

    Returns
    -------
    :class:`bool`
        Is specification a grey colour.

    Examples
    --------
    >>> is_grey_munsell_colour(np.array([0.0, 2.0, 4.0, 6]))
    False
    >>> is_grey_munsell_colour(np.array([np.nan, 0.5, np.nan, np.nan]))
    True
    """

    specification = as_float_array(specification)

    specification = specification[~np.isnan(specification)]

    return is_numeric(as_float(specification))


def normalise_munsell_specification(specification: ArrayLike) -> NDArray:
    """
    Normalise given *Munsell* *Colorlab* specification.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.

    Returns
    -------
    :class:`numpy.ndarray`
        Normalised *Munsell* *Colorlab* specification.

    Examples
    --------
    >>> normalise_munsell_specification(
    ...     np.array([0.0, 2.0, 4.0, 6]))
    array([ 10.,   2.,   4.,   7.])
    >>> normalise_munsell_specification(
    ...     np.array([np.nan, 0.5, np.nan, np.nan]))
    array([ nan,  0.5,  nan,  nan])
    """

    specification = as_float_array(specification)

    if is_grey_munsell_colour(specification):
        return specification * np.array([np.nan, 1, np.nan, np.nan])
    else:
        hue, value, chroma, code = specification

        if hue == 0:
            # 0YR is equivalent to 10R.
            hue, code = 10, (code + 1) % 10

        if chroma == 0:
            return tstack([np.nan, value, np.nan, np.nan])
        else:
            return tstack([hue, value, chroma, code])


def munsell_colour_to_munsell_specification(munsell_colour: str) -> NDArray:
    """
    Retrieve a normalised *Munsell* *Colorlab* specification from given
    *Munsell* colour.

    Parameters
    ----------
    munsell_colour
        *Munsell* colour.

    Returns
    -------
    :class:`numpy.ndarray`
        Normalised *Munsell* *Colorlab* specification.

    Examples
    --------
    >>> munsell_colour_to_munsell_specification('N5.2')
    array([ nan,  5.2,  nan,  nan])
    >>> munsell_colour_to_munsell_specification('0YR 2.0/4.0')
    array([ 10.,   2.,   4.,   7.])
    """

    return normalise_munsell_specification(
        parse_munsell_colour(munsell_colour)
    )


def munsell_specification_to_munsell_colour(
    specification: ArrayLike,
    hue_decimals: Integer = 1,
    value_decimals: Integer = 1,
    chroma_decimals: Integer = 1,
) -> str:
    """
    Convert from *Munsell* *Colorlab* specification to given *Munsell* colour.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.
    hue_decimals
        Hue formatting decimals.
    value_decimals
        Value formatting decimals.
    chroma_decimals
        Chroma formatting decimals.

    Returns
    -------
    :class:`str`
        *Munsell* colour.

    Examples
    --------
    >>> munsell_specification_to_munsell_colour(
    ...     np.array([np.nan, 5.2, np.nan, np.nan]))
    'N5.2'
    >>> munsell_specification_to_munsell_colour(
    ...     np.array([10, 2.0, 4.0, 7]))
    '10.0R 2.0/4.0'
    """

    hue, value, chroma, code = tsplit(
        normalise_munsell_specification(specification)
    )

    if is_grey_munsell_colour(specification):
        return MUNSELL_GRAY_EXTENDED_FORMAT.format(value, value_decimals)
    else:
        hue = round(hue, hue_decimals)
        attest(
            0 <= hue <= 10,
            f'"{specification!r}" specification hue must be normalised to '
            f"domain [0, 10]!",
        )

        value = round(value, value_decimals)
        attest(
            0 <= value <= 10,
            f'"{specification!r}" specification value must be normalised to '
            f"domain [0, 10]!",
        )

        chroma = round(chroma, chroma_decimals)
        attest(
            2 <= chroma <= 50,
            f'"{specification!r}" specification chroma must be normalised to '
            f"domain [2, 50]!",
        )

        code_values = MUNSELL_HUE_LETTER_CODES.values()
        code = round(code, 1)
        attest(
            code in code_values,
            f'"{specification!r}" specification code must one of '
            f'"{code_values}"!',
        )

        if value == 0:
            return MUNSELL_GRAY_EXTENDED_FORMAT.format(value, value_decimals)
        else:
            hue_letter = MUNSELL_HUE_LETTER_CODES.first_key_from_value(code)

            return MUNSELL_COLOUR_EXTENDED_FORMAT.format(
                hue,
                hue_decimals,
                hue_letter,
                value,
                value_decimals,
                chroma,
                chroma_decimals,
            )


def xyY_from_renotation(specification: ArrayLike) -> NDArray:
    """
    Return given existing *Munsell* *Colorlab* specification *CIE xyY*
    colourspace vector from *Munsell Renotation System* data.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xyY* colourspace vector.

    Raises
    ------
    ValueError
        If the given specification doesn't exist in *Munsell Renotation System*
        data.

    Examples
    --------
    >>> xyY_from_renotation(np.array([2.5, 0.2, 2.0, 4]))  # doctest: +ELLIPSIS
    array([ 0.71...,  1.41...,  0.23...])
    """

    specification = normalise_munsell_specification(specification)

    try:
        index = np.where(
            (_munsell_specifications() == specification).all(axis=-1)
        )

        return MUNSELL_COLOURS_ALL[int(index[0])][1]
    except Exception:
        raise ValueError(
            f'"{specification}" specification does not exists in '
            '"Munsell Renotation System" data!'
        )


def is_specification_in_renotation(specification: ArrayLike) -> Boolean:
    """
    Return whether given *Munsell* *Colorlab* specification is in
    *Munsell Renotation System* data.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.

    Returns
    -------
    :class:`bool`
        Whether specification is in *Munsell Renotation System* data.

    Examples
    --------
    >>> is_specification_in_renotation(np.array([2.5, 0.2, 2.0, 4]))
    True
    >>> is_specification_in_renotation(np.array([64, 0.2, 2.0, 4]))
    False
    """

    try:
        xyY_from_renotation(specification)

        return True
    except ValueError:
        return False


def bounding_hues_from_renotation(hue_and_code: ArrayLike) -> NDArray:
    """
    Return for a given *Munsell* *Colorlab* specification hue and *Munsell*
    *Colorlab* specification code the two bounding hues from
    *Munsell Renotation System* data.

    Parameters
    ----------
    hue_and_code
        *Munsell* *Colorlab* specification hue and *Munsell* *Colorlab*
        specification code.

    Returns
    -------
    :class:`numpy.ndarray`
        Bounding hues.

    References
    ----------
    :cite:`Centore2014o`

    Examples
    --------
    >>> bounding_hues_from_renotation([3.2, 4])
    array([[ 2.5,  4. ],
           [ 5. ,  4. ]])

    # Coverage Doctests

    >>> bounding_hues_from_renotation([0.0, 1])
    array([[ 10.,   2.],
           [ 10.,   2.]])
    """

    hue, code = as_float_array(hue_and_code)

    hue_cw: Floating
    code_cw: Floating
    hue_ccw: Floating
    code_ccw: Floating

    if hue % 2.5 == 0:
        if hue == 0:
            hue_cw = 10
            code_cw = (code + 1) % 10
        else:
            hue_cw = hue
            code_cw = code
        hue_ccw = hue_cw
        code_ccw = code_cw
    else:
        hue_cw = 2.5 * np.floor(hue / 2.5)
        hue_ccw = (hue_cw + 2.5) % 10
        if hue_ccw == 0:
            hue_ccw = 10

        if hue_cw == 0:
            hue_cw = 10
            code_cw = (code + 1) % 10
            if code_cw == 0:
                code_cw = 10
        else:
            code_cw = code
        code_ccw = code

    return as_float_array([(hue_cw, code_cw), (hue_ccw, code_ccw)])


def hue_to_hue_angle(hue_and_code: ArrayLike) -> Floating:
    """
    Convert from the *Munsell* *Colorlab* specification hue and *Munsell*
    *Colorlab* specification code to hue angle in degrees.

    Parameters
    ----------
    hue_and_code
        *Munsell* *Colorlab* specification hue and *Munsell* *Colorlab*
        specification code.

    Returns
    -------
    :class:`numpy.floating`
        Hue angle in degrees.

    References
    ----------
    :cite:`Centore2014s`

    Examples
    --------
    >>> hue_to_hue_angle([3.2, 4])
    65.5
    """

    hue, code = as_float_array(hue_and_code)

    single_hue = ((17 - code) % 10 + (hue / 10) - 0.5) % 10

    hue_angle = LinearInterpolator(
        [0, 2, 3, 4, 5, 6, 8, 9, 10], [0, 45, 70, 135, 160, 225, 255, 315, 360]
    )(single_hue)

    return as_float_scalar(hue_angle)


def hue_angle_to_hue(hue_angle: Floating) -> NDArray:
    """
    Convert from hue angle in degrees to the *Munsell* *Colorlab*
    specification hue and code.

    Parameters
    ----------
    hue_angle
        Hue angle in degrees.

    Returns
    -------
    :class:`numpy.ndarray`
        (*Munsell* *Colorlab* specification hue, *Munsell* *Colorlab*
        specification code).

    References
    ----------
    :cite:`Centore2014t`

    Examples
    --------
    >>> hue_angle_to_hue(65.54)  # doctest: +ELLIPSIS
    array([ 3.216,  4.   ])
    """

    single_hue = LinearInterpolator(
        [0, 45, 70, 135, 160, 225, 255, 315, 360], [0, 2, 3, 4, 5, 6, 8, 9, 10]
    )(hue_angle)

    if single_hue <= 0.5:
        code = 7
    elif single_hue <= 1.5:
        code = 6
    elif single_hue <= 2.5:
        code = 5
    elif single_hue <= 3.5:
        code = 4
    elif single_hue <= 4.5:
        code = 3
    elif single_hue <= 5.5:
        code = 2
    elif single_hue <= 6.5:
        code = 1
    elif single_hue <= 7.5:
        code = 10
    elif single_hue <= 8.5:
        code = 9
    elif single_hue <= 9.5:
        code = 8
    else:
        code = 7

    hue = (10 * (single_hue % 1) + 5) % 10
    if hue == 0:
        hue = 10

    return tstack([hue, code])


def hue_to_ASTM_hue(hue_and_code) -> Floating:
    """
    Convert from the *Munsell* *Colorlab* specification hue and *Munsell*
    *Colorlab* specification codeto *ASTM* hue number.

    Parameters
    ----------
    hue_and_code
        *Munsell* *Colorlab* specification hue and *Munsell* *Colorlab*
        specification code.


    Returns
    -------
    :class:`numpy.floating`
        *ASTM* hue number.

    References
    ----------
    :cite:`Centore2014k`

    Examples
    --------
    >>> hue_to_ASTM_hue([3.2, 4])  # doctest: +ELLIPSIS
    33.2...
    """

    hue, code = as_float_array(hue_and_code)

    ASTM_hue = 10 * ((7 - code) % 10) + hue

    return 100 if ASTM_hue == 0 else ASTM_hue


def interpolation_method_from_renotation_ovoid(
    specification: ArrayLike,
) -> Optional[Literal["Linear", "Radial"]]:
    """
    Return whether to use linear or radial interpolation when drawing ovoids
    through data points in the *Munsell Renotation System* data from given
    specification.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.

    Returns
    -------
    :py:data:`None` or :class:`str`
        Interpolation method.

    References
    ----------
    :cite:`Centore2014l`

    Examples
    --------
    >>> interpolation_method_from_renotation_ovoid([2.5, 5.0, 12.0, 4])
    'Radial'
    """

    specification = normalise_munsell_specification(specification)

    interpolation_methods: Dict[
        Integer, Optional[Literal["Linear", "Radial"]]
    ] = {
        0: None,
        1: "Linear",
        2: "Radial",
    }

    if is_grey_munsell_colour(specification):
        # No interpolation needed for grey colours.
        interpolation_method = 0
    else:
        hue, value, chroma, code = specification

        attest(
            0 <= value <= 10,
            f'"{specification}" specification value must be normalised to '
            f"domain [0, 10]!",
        )

        attest(
            is_integer(value),
            f'"{specification}" specification value must be an integer!',
        )

        value = round(value)

        attest(
            2 <= chroma <= 50,
            f'"{specification}" specification chroma must be normalised to '
            f"domain [2, 50]!",
        )

        attest(
            abs(2 * (chroma / 2 - round(chroma / 2))) <= INTEGER_THRESHOLD,
            f'"{specification}" specification chroma must be an integer and '
            f"multiple of 2!",
        )

        chroma = 2 * round(chroma / 2)

        interpolation_method = 0

        # Standard Munsell Renotation System hue, no interpolation needed.
        if hue % 2.5 == 0:
            interpolation_method = 0

        ASTM_hue = hue_to_ASTM_hue([hue, code])

        if value == 1:
            if chroma == 2:
                if 15 < ASTM_hue < 30 or 60 < ASTM_hue < 85:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 4:
                if 12.5 < ASTM_hue < 27.5 or 57.5 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 6:
                if 55 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 8:
                if 67.5 < ASTM_hue < 77.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma >= 10:
                # NOTE: This condition is likely never "True" while producing a
                # valid "Munsell Specification" in practice: 1M iterations with
                # random numbers never reached this code path while producing a
                # valid "Munsell Specification".
                if 72.5 < ASTM_hue < 77.5:  # pragma: no cover
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            else:  # pragma: no cover
                interpolation_method = 1
        elif value == 2:
            if chroma == 2:
                if 15 < ASTM_hue < 27.5 or 77.5 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 4:
                if 12.5 < ASTM_hue < 30 or 62.5 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 6:
                if 7.5 < ASTM_hue < 22.5 or 62.5 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 8:
                if 7.5 < ASTM_hue < 15 or 60 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma >= 10:
                if 65 < ASTM_hue < 77.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            else:  # pragma: no cover
                interpolation_method = 1
        elif value == 3:
            if chroma == 2:
                if 10 < ASTM_hue < 37.5 or 65 < ASTM_hue < 85:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 4:
                if 5 < ASTM_hue < 37.5 or 55 < ASTM_hue < 72.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma in (6, 8, 10):
                if 7.5 < ASTM_hue < 37.5 or 57.5 < ASTM_hue < 82.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma >= 12:
                if 7.5 < ASTM_hue < 42.5 or 57.5 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            else:  # pragma: no cover
                interpolation_method = 1
        elif value == 4:
            if chroma in (2, 4):
                if 7.5 < ASTM_hue < 42.5 or 57.5 < ASTM_hue < 85:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma in (6, 8):
                if 7.5 < ASTM_hue < 40 or 57.5 < ASTM_hue < 82.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma >= 10:
                if 7.5 < ASTM_hue < 40 or 57.5 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            else:  # pragma: no cover
                interpolation_method = 1
        elif value == 5:
            if chroma == 2:
                if 5 < ASTM_hue < 37.5 or 55 < ASTM_hue < 85:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma in (4, 6, 8):
                if 2.5 < ASTM_hue < 42.5 or 55 < ASTM_hue < 85:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma >= 10:
                if 2.5 < ASTM_hue < 42.5 or 55 < ASTM_hue < 82.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            else:  # pragma: no cover
                interpolation_method = 1
        elif value == 6:
            if chroma in (2, 4):
                if 5 < ASTM_hue < 37.5 or 55 < ASTM_hue < 87.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 6:
                if 5 < ASTM_hue < 42.5 or 57.5 < ASTM_hue < 87.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma in (8, 10):
                if 5 < ASTM_hue < 42.5 or 60 < ASTM_hue < 85:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma in (12, 14):
                if 5 < ASTM_hue < 42.5 or 60 < ASTM_hue < 82.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma >= 16:
                if 5 < ASTM_hue < 42.5 or 60 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            else:  # pragma: no cover
                interpolation_method = 1
        elif value == 7:
            if chroma in (2, 4, 6):
                if 5 < ASTM_hue < 42.5 or 60 < ASTM_hue < 85:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 8:
                if 5 < ASTM_hue < 42.5 or 60 < ASTM_hue < 82.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 10:
                if (
                    30 < ASTM_hue < 42.5
                    or 5 < ASTM_hue < 25
                    or 60 < ASTM_hue < 82.5
                ):
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma == 12:
                if (
                    30 < ASTM_hue < 42.5
                    or 7.5 < ASTM_hue < 27.5
                    or 80 < ASTM_hue < 82.5
                ):
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma >= 14:
                if (
                    32.5 < ASTM_hue < 40
                    or 7.5 < ASTM_hue < 15
                    or 80 < ASTM_hue < 82.5
                ):
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            else:  # pragma: no cover
                interpolation_method = 1
        elif value == 8:
            if chroma in (2, 4, 6, 8, 10, 12):
                if 5 < ASTM_hue < 40 or 60 < ASTM_hue < 85:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma >= 14:
                if (
                    32.5 < ASTM_hue < 40
                    or 5 < ASTM_hue < 15
                    or 60 < ASTM_hue < 85
                ):
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            else:  # pragma: no cover
                interpolation_method = 1
        elif value == 9:
            if chroma in (2, 4):
                if 5 < ASTM_hue < 40 or 55 < ASTM_hue < 80:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma in (6, 8, 10, 12, 14):
                if 5 < ASTM_hue < 42.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            elif chroma >= 16:
                if 35 < ASTM_hue < 42.5:
                    interpolation_method = 2
                else:
                    interpolation_method = 1
            else:  # pragma: no cover
                interpolation_method = 1
        elif value == 10:
            # Ideal white, no interpolation needed.
            interpolation_method = 0

    return interpolation_methods[interpolation_method]


def xy_from_renotation_ovoid(specification: ArrayLike) -> NDArray:
    """
    Convert given *Munsell* *Colorlab* specification to *CIE xy* chromaticity
    coordinates on *Munsell Renotation System* ovoid.
    The *CIE xy* point will be on the ovoid about the achromatic point,
    corresponding to the *Munsell* *Colorlab* specification
    value and chroma.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xy* chromaticity coordinates.

    Raises
    ------
    ValueError
        If an invalid interpolation method is retrieved from internal
        computations.

    References
    ----------
    :cite:`Centore2014n`

    Examples
    --------
    >>> xy_from_renotation_ovoid([2.5, 5.0, 12.0, 4])
    ... # doctest: +ELLIPSIS
    array([ 0.4333...,  0.5602...])
    >>> xy_from_renotation_ovoid([np.nan, 8, np.nan, np.nan])
    ... # doctest: +ELLIPSIS
    array([ 0.31006...,  0.31616...])
    """

    specification = normalise_munsell_specification(specification)

    if is_grey_munsell_colour(specification):
        return CCS_ILLUMINANT_MUNSELL
    else:
        hue, value, chroma, code = specification

        attest(
            1 <= value <= 9,
            f'"{specification}" specification value must be normalised to '
            f"domain [1, 9]!",
        )

        attest(
            is_integer(value),
            f'"{specification}" specification value must be an integer!',
        )

        value = round(value)

        attest(
            2 <= chroma <= 50,
            f'"{specification}" specification chroma must be normalised to '
            f"domain [2, 50]!",
        )

        attest(
            abs(2 * (chroma / 2 - round(chroma / 2))) <= INTEGER_THRESHOLD,
            f'"{specification}" specification chroma must be an integer and '
            f"multiple of 2!",
        )

        chroma = 2 * round(chroma / 2)

        # Checking if renotation data is available without interpolation using
        # given threshold.
        threshold = 1e-7
        if (
            abs(hue) < threshold
            or abs(hue - 2.5) < threshold
            or abs(hue - 5) < threshold
            or abs(hue - 7.5) < threshold
            or abs(hue - 10) < threshold
        ):
            hue = 2.5 * round(hue / 2.5)

            x, y, _Y = xyY_from_renotation([hue, value, chroma, code])

            return tstack([x, y])

        hue_cw, hue_ccw = bounding_hues_from_renotation([hue, code])
        hue_minus, code_minus = hue_cw
        hue_plus, code_plus = hue_ccw

        x_grey, y_grey = CCS_ILLUMINANT_MUNSELL

        specification_minus = (hue_minus, value, chroma, code_minus)
        x_minus, y_minus, Y_minus = xyY_from_renotation(specification_minus)
        rho_minus, phi_minus, _z_minus = cartesian_to_cylindrical(
            [x_minus - x_grey, y_minus - y_grey, Y_minus]
        )
        phi_minus = np.degrees(phi_minus)

        specification_plus = (hue_plus, value, chroma, code_plus)
        x_plus, y_plus, Y_plus = xyY_from_renotation(specification_plus)
        rho_plus, phi_plus, _z_plus = cartesian_to_cylindrical(
            [x_plus - x_grey, y_plus - y_grey, Y_plus]
        )
        phi_plus = np.degrees(phi_plus)

        hue_angle_lower = hue_to_hue_angle([hue_minus, code_minus])
        hue_angle = hue_to_hue_angle([hue, code])
        hue_angle_upper = hue_to_hue_angle([hue_plus, code_plus])

        if phi_minus - phi_plus > 180:
            phi_plus += 360

        if hue_angle_lower == 0:
            hue_angle_lower = 360

        if hue_angle_lower > hue_angle_upper:
            if hue_angle_lower > hue_angle:
                hue_angle_lower -= 360
            else:
                hue_angle_lower -= 360
                hue_angle -= 360

        interpolation_method = interpolation_method_from_renotation_ovoid(
            specification
        )

        attest(
            interpolation_method is not None,
            f"Interpolation method must be one of: "
            f"\"{', '.join(['Linear', 'Radial'])}\"",
        )

        hue_angle_lower_upper = np.squeeze(
            as_float_array([hue_angle_lower, hue_angle_upper])
        )

        if interpolation_method == "Linear":
            x_minus_plus = np.squeeze([x_minus, x_plus])
            y_minus_plus = np.squeeze([y_minus, y_plus])

            x = LinearInterpolator(hue_angle_lower_upper, x_minus_plus)(
                hue_angle
            )
            y = LinearInterpolator(hue_angle_lower_upper, y_minus_plus)(
                hue_angle
            )
        elif interpolation_method == "Radial":
            rho_minus_plus = np.squeeze([rho_minus, rho_plus])
            phi_minus_plus = np.squeeze([phi_minus, phi_plus])

            rho = as_float_array(
                LinearInterpolator(hue_angle_lower_upper, rho_minus_plus)(
                    hue_angle
                )
            )
            phi = as_float_array(
                LinearInterpolator(hue_angle_lower_upper, phi_minus_plus)(
                    hue_angle
                )
            )

            rho_phi = np.squeeze([rho, np.radians(phi)])
            x, y = tsplit(
                polar_to_cartesian(rho_phi) + tstack([x_grey, y_grey])
            )

        return tstack([x, y])


def LCHab_to_munsell_specification(LCHab: ArrayLike) -> NDArray:
    """
    Convert from *CIE L\\*C\\*Hab* colourspace to approximate *Munsell*
    *Colorlab* specification.

    Parameters
    ----------
    LCHab
        *CIE L\\*C\\*Hab* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *Munsell* *Colorlab* specification.

    References
    ----------
    :cite:`Centore2014u`

    Examples
    --------
    >>> LCHab = np.array([100, 17.50664796, 244.93046842])
    >>> LCHab_to_munsell_specification(LCHab)  # doctest: +ELLIPSIS
    array([  8.0362412...,  10.        ,   3.5013295...,   1.        ])
    """

    L, C, Hab = tsplit(LCHab)

    if Hab == 0:
        code = 8
    elif Hab <= 36:
        code = 7
    elif Hab <= 72:
        code = 6
    elif Hab <= 108:
        code = 5
    elif Hab <= 144:
        code = 4
    elif Hab <= 180:
        code = 3
    elif Hab <= 216:
        code = 2
    elif Hab <= 252:
        code = 1
    elif Hab <= 288:
        code = 10
    elif Hab <= 324:
        code = 9
    else:
        code = 8

    hue = LinearInterpolator([0, 36], [0, 10])(Hab % 36)
    if hue == 0:
        hue = 10

    value = L / 10
    chroma = C / 5

    return tstack([hue, value, chroma, code])


def maximum_chroma_from_renotation(
    hue_and_value_and_code: ArrayLike,
) -> Floating:
    """
    Return the maximum *Munsell* chroma from *Munsell Renotation System* data
    using given *Munsell* *Colorlab* specification hue, *Munsell* *Colorlab*
    specification value and *Munsell* *Colorlab* specification code.

    Parameters
    ----------
    hue_and_value_and_code
        *Munsell* *Colorlab* specification hue, *Munsell* *Colorlab*
        specification value and *Munsell* *Colorlab* specification code.

    Returns
    -------
    :class:`numpy.floating`
        Maximum chroma.

    References
    ----------
    :cite:`Centore2014r`

    Examples
    --------
    >>> maximum_chroma_from_renotation([2.5, 5, 5])
    14.0
    """

    hue, value, code = as_float_array(hue_and_value_and_code)

    # Ideal white, no chroma.
    if value >= 9.99:
        return 0

    attest(
        1 <= value <= 10,
        f'"{value}" value must be normalised to domain [1, 10]!',
    )

    if value % 1 == 0:
        value_minus = value
        value_plus = value
    else:
        value_minus = np.floor(value)
        value_plus = value_minus + 1

    hue_cw, hue_ccw = bounding_hues_from_renotation([hue, code])
    hue_cw, code_cw = hue_cw
    hue_ccw, code_ccw = hue_ccw

    maximum_chromas = _munsell_maximum_chromas_from_renotation()
    specification_for_indexes = [chroma[0] for chroma in maximum_chromas]

    ma_limit_mcw = maximum_chromas[
        specification_for_indexes.index((hue_cw, value_minus, code_cw))
    ][1]
    ma_limit_mccw = maximum_chromas[
        specification_for_indexes.index((hue_ccw, value_minus, code_ccw))
    ][1]

    if value_plus <= 9:
        ma_limit_pcw = maximum_chromas[
            specification_for_indexes.index((hue_cw, value_plus, code_cw))
        ][1]
        ma_limit_pccw = maximum_chromas[
            specification_for_indexes.index((hue_ccw, value_plus, code_ccw))
        ][1]
        max_chroma = min(
            [ma_limit_mcw, ma_limit_mccw, ma_limit_pcw, ma_limit_pccw]
        )
    else:
        L = as_float_scalar(luminance_ASTMD1535(value))
        L9 = as_float_scalar(luminance_ASTMD1535(9))
        L10 = as_float_scalar(luminance_ASTMD1535(10))

        max_chroma = min(
            [
                as_float_scalar(
                    LinearInterpolator([L9, L10], [ma_limit_mcw, 0])(L)
                ),
                as_float_scalar(
                    LinearInterpolator([L9, L10], [ma_limit_mccw, 0])(L)
                ),
            ]
        )

    return max_chroma


def munsell_specification_to_xy(specification: ArrayLike) -> NDArray:
    """
    Convert given *Munsell* *Colorlab* specification to *CIE xy* chromaticity
    coordinates by interpolating over *Munsell Renotation System* data.

    Parameters
    ----------
    specification
        *Munsell* *Colorlab* specification.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xy* chromaticity coordinates.

    References
    ----------
    :cite:`Centore2014q`

    Examples
    --------
    >>> munsell_specification_to_xy([2.1, 8.0, 17.9, 4])
    ... # doctest: +ELLIPSIS
    array([ 0.4400632...,  0.5522428...])
    >>> munsell_specification_to_xy([np.nan, 8, np.nan, np.nan])
    ... # doctest: +ELLIPSIS
    array([ 0.31006...,  0.31616...])
    """

    specification = normalise_munsell_specification(specification)

    if is_grey_munsell_colour(specification):
        return CCS_ILLUMINANT_MUNSELL
    else:
        hue, value, chroma, code = specification

        attest(
            0 <= value <= 10,
            f'"{specification}" specification value must be normalised to '
            f"domain [0, 10]!",
        )

        attest(
            is_integer(value),
            f'"{specification}" specification value must be an integer!',
        )

        value = round(value)

        if chroma % 2 == 0:
            chroma_minus = chroma_plus = chroma
        else:
            chroma_minus = 2 * np.floor(chroma / 2)
            chroma_plus = chroma_minus + 2

        if chroma_minus == 0:
            # Smallest chroma ovoid collapses to illuminant chromaticity
            # coordinates.
            x_minus, y_minus = CCS_ILLUMINANT_MUNSELL
        else:
            x_minus, y_minus = xy_from_renotation_ovoid(
                [hue, value, chroma_minus, code]
            )

        x_plus, y_plus = xy_from_renotation_ovoid(
            [hue, value, chroma_plus, code]
        )

        if chroma_minus == chroma_plus:
            x = x_minus
            y = y_minus
        else:
            chroma_minus_plus = np.squeeze([chroma_minus, chroma_plus])
            x_minus_plus = np.squeeze([x_minus, x_plus])
            y_minus_plus = np.squeeze([y_minus, y_plus])

            x = LinearInterpolator(chroma_minus_plus, x_minus_plus)(chroma)
            y = LinearInterpolator(chroma_minus_plus, y_minus_plus)(chroma)

        return tstack([x, y])
